# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mutual Information Gap from the beta-TC-VAE paper.

Based on "Isolating Sources of Disentanglement in Variational Autoencoders"
(https://arxiv.org/pdf/1802.04942.pdf).
"""
from absl import logging
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
import gin.tf


@gin.configurable(
    "mig",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_mig(ground_truth_data,
                representation_function,
                random_state,
                artifact_dir=None,
                num_train=gin.REQUIRED,
                batch_size=16):
  """Computes the mutual information gap.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    batch_size: Batch size for sampling.

  Returns:
    Dict with average mutual information gap.
  """
  del artifact_dir
  logging.info("Generating training set.")
  mus_train, ys_train = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train,
      random_state, batch_size)
  assert mus_train.shape[1] == num_train
  return _compute_mig(mus_train, ys_train)


def _compute_mig(mus_train, ys_train):
  """Computes score based on both training and testing codes and factors."""
  score_dict = {}
  discretized_mus = utils.make_discretizer(mus_train)
  m = utils.discrete_mutual_info(discretized_mus, ys_train)
  assert m.shape[0] == mus_train.shape[0]
  assert m.shape[1] == ys_train.shape[0]
  # m is [num_latents, num_factors]
  entropy = utils.discrete_entropy(ys_train)
  sorted_m = np.sort(m, axis=0)[::-1]
  score_dict["discrete_mig"] = np.mean(
      np.divide(sorted_m[0, :] - sorted_m[1, :], entropy[:]))
  return score_dict


@gin.configurable(
    "mig_validation",
    blacklist=["observations", "labels", "representation_function"])
def compute_mig_on_fixed_data(observations, labels, representation_function,
                              batch_size=100):
  """Computes the MIG scores on the fixed set of observations and labels.

  Args:
    observations: Observations on which to compute the score. Observations have
      shape (num_observations, 64, 64, num_channels).
    labels: Observed factors of variations.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    batch_size: Batch size used to compute the representation.

  Returns:
    MIG computed on the provided observations and labels.
  """
  mus = utils.obtain_representation(observations, representation_function,
                                    batch_size)
  assert labels.shape[1] == observations.shape[0], "Wrong labels shape."
  assert mus.shape[1] == observations.shape[0], "Wrong representation shape."
  return _compute_mig(mus, labels)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for convolute_hub.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.utils import convolute_hub
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub


class ConvoluteHubTest(tf.test.TestCase):

  def test_convolute(self):

    # Create variables to use.
    random_state = np.random.RandomState(0)
    data = random_state.normal(size=(5, 10))
    variable1 = random_state.normal(size=(10, 6))
    variable2 = random_state.normal(size=(6, 2))

    # Save variables to checkpoint.
    checkpoint_path = os.path.join(self.get_temp_dir(), "checkpoint.ckpt")
    convolute_hub.save_numpy_arrays_to_checkpoint(
        checkpoint_path, variable1=variable1, variable2=variable2)

    # Save a TFHub module that we will convolute.
    module_path = os.path.join(self.get_temp_dir(), "module_path")
    def module_fn():
      tensor = tf.placeholder(tf.float64, shape=(None, 10))
      variable1 = tf.get_variable("variable1", shape=(10, 6), dtype=tf.float64)
      output = tf.matmul(tensor, variable1)
      hub.add_signature(
          name="multiplication1",
          inputs={"tensor": tensor},
          outputs={"tensor": output})
    spec = hub.create_module_spec(module_fn)
    spec.export(module_path, checkpoint_path=checkpoint_path)

    # Function used for the convolution.
    def _operation2(tensor):
      variable2 = tf.get_variable("variable2", shape=(6, 2), dtype=tf.float64)
      return dict(tensor=tf.matmul(tensor, variable2))

    # Save the convolution as a new TFHub module
    module_path_new = os.path.join(self.get_temp_dir(), "module_path_new")
    convolute_hub.convolute_and_save(module_path, "multiplication1",
                                     module_path_new, _operation2,
                                     checkpoint_path, "convoluted")

    # Check the first signature.
    with hub.eval_function_for_module(module_path_new) as f:
      module_result = f(dict(tensor=data), signature="convoluted", as_dict=True)
    real_result = data.dot(variable1).dot(variable2)
    self.assertAllClose(real_result, module_result["tensor"])


if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Modularity and explicitness metrics from the F-statistic paper.

Based on "Learning Deep Disentangled Embeddings With the F-Statistic Loss"
(https://arxiv.org/pdf/1802.05312.pdf).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
from sklearn import linear_model
from sklearn import metrics
from sklearn import preprocessing
import gin.tf


@gin.configurable(
    "modularity_explicitness",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_modularity_explicitness(ground_truth_data,
                                    representation_function,
                                    random_state,
                                    artifact_dir=None,
                                    num_train=gin.REQUIRED,
                                    num_test=gin.REQUIRED,
                                    batch_size=16):
  """Computes the modularity metric according to Sec 3.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    batch_size: Batch size for sampling.

  Returns:
    Dictionary with average modularity score and average explicitness
      (train and test).
  """
  del artifact_dir
  scores = {}
  mus_train, ys_train = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train,
      random_state, batch_size)
  mus_test, ys_test = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_test,
      random_state, batch_size)
  discretized_mus = utils.make_discretizer(mus_train)
  mutual_information = utils.discrete_mutual_info(discretized_mus, ys_train)
  # Mutual information should have shape [num_codes, num_factors].
  assert mutual_information.shape[0] == mus_train.shape[0]
  assert mutual_information.shape[1] == ys_train.shape[0]
  scores["modularity_score"] = modularity(mutual_information)
  explicitness_score_train = np.zeros([ys_train.shape[0], 1])
  explicitness_score_test = np.zeros([ys_test.shape[0], 1])
  mus_train_norm, mean_mus, stddev_mus = utils.normalize_data(mus_train)
  mus_test_norm, _, _ = utils.normalize_data(mus_test, mean_mus, stddev_mus)
  for i in range(ys_train.shape[0]):
    explicitness_score_train[i], explicitness_score_test[i] = \
        explicitness_per_factor(mus_train_norm, ys_train[i, :],
                                mus_test_norm, ys_test[i, :])
  scores["explicitness_score_train"] = np.mean(explicitness_score_train)
  scores["explicitness_score_test"] = np.mean(explicitness_score_test)
  return scores


def explicitness_per_factor(mus_train, y_train, mus_test, y_test):
  """Compute explicitness score for a factor as ROC-AUC of a classifier.

  Args:
    mus_train: Representation for training, (num_codes, num_points)-np array.
    y_train: Ground truth factors for training, (num_factors, num_points)-np
      array.
    mus_test: Representation for testing, (num_codes, num_points)-np array.
    y_test: Ground truth factors for testing, (num_factors, num_points)-np
      array.

  Returns:
    roc_train: ROC-AUC score of the classifier on training data.
    roc_test: ROC-AUC score of the classifier on testing data.
  """
  x_train = np.transpose(mus_train)
  x_test = np.transpose(mus_test)
  clf = linear_model.LogisticRegression().fit(x_train, y_train)
  y_pred_train = clf.predict_proba(x_train)
  y_pred_test = clf.predict_proba(x_test)
  mlb = preprocessing.MultiLabelBinarizer()
  roc_train = metrics.roc_auc_score(
      mlb.fit_transform(np.expand_dims(y_train, 1)), y_pred_train)
  roc_test = metrics.roc_auc_score(
      mlb.fit_transform(np.expand_dims(y_test, 1)), y_pred_test)
  return roc_train, roc_test


def modularity(mutual_information):
  """Computes the modularity from mutual information."""
  # Mutual information has shape [num_codes, num_factors].
  squared_mi = np.square(mutual_information)
  max_squared_mi = np.max(squared_mi, axis=1)
  numerator = np.sum(squared_mi, axis=1) - max_squared_mi
  denominator = max_squared_mi * (squared_mi.shape[1] -1.)
  delta = numerator / denominator
  modularity_score = 1. - delta
  index = (max_squared_mi == 0.)
  modularity_score[index] = 0.
  return np.mean(modularity_score)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shapes3D data set."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.data.ground_truth import ground_truth_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
from six.moves import range
import tensorflow.compat.v1 as tf


SHAPES3D_PATH = os.path.join(
    os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "3dshapes",
    "look-at-object-room_floor-hueXwall-hueXobj-"
    "hueXobj-sizeXobj-shapeXview-azi.npz"
)



class Shapes3D(ground_truth_data.GroundTruthData):
  """Shapes3D dataset.

  The data set was originally introduced in "Disentangling by Factorising".

  The ground-truth factors of variation are:
  0 - floor color (10 different values)
  1 - wall color (10 different values)
  2 - object color (10 different values)
  3 - object size (8 different values)
  4 - object type (4 different values)
  5 - azimuth (15 different values)
  """

  def __init__(self):
    with tf.gfile.GFile(SHAPES3D_PATH, "rb") as f:
      # Data was saved originally using python2, so we need to set the encoding.
      data = np.load(f, encoding="latin1")
    images = data["images"]
    labels = data["labels"]
    n_samples = np.prod(images.shape[0:6])
    self.images = (
        images.reshape([n_samples, 64, 64, 3]).astype(np.float32) / 255.)
    features = labels.reshape([n_samples, 6])
    self.factor_sizes = [10, 10, 10, 8, 4, 15]
    self.latent_factor_indices = list(range(6))
    self.num_total_factors = features.shape[1]
    self.state_space = util.SplitDiscreteStateSpace(self.factor_sizes,
                                                    self.latent_factor_indices)
    self.factor_bases = np.prod(self.factor_sizes) / np.cumprod(
        self.factor_sizes)

  @property
  def num_factors(self):
    return self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return self.factor_sizes

  @property
  def observation_shape(self):
    return [64, 64, 3]


  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return self.state_space.sample_latent_factors(num, random_state)

  def sample_observations_from_factors(self, factors, random_state):
    all_factors = self.state_space.sample_all_factors(factors, random_state)
    indices = np.array(np.dot(all_factors, self.factor_bases), dtype=np.int64)
    return self.images[indices]



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization code for Interventional Robustness Score.

Based on the paper https://arxiv.org/abs/1811.00007.
"""
import os

from disentanglement_lib.evaluation.metrics.irs import scalable_disentanglement_score
import matplotlib
matplotlib.use("Agg")  # Set headless-friendly backend.
import matplotlib.pyplot as plt  # pylint: disable=g-import-not-at-top
import numpy as np
import tensorflow.compat.v1 as tf


def vis_all_interventional_effects(gen_factors, latents, output_dir):
  """Compute Matrix of all interventional effects."""
  res = scalable_disentanglement_score(gen_factors, latents)
  parents = res["parents"]
  scores = res["disentanglement_scores"]

  fig_width_inches = 3.0 * gen_factors.shape[1]
  fig_height_inches = 3.0 * latents.shape[1]
  fig, axes = plt.subplots(
      latents.shape[1],
      gen_factors.shape[1],
      figsize=(fig_width_inches, fig_height_inches),
      sharex="col",
      sharey="row")

  for j in range(gen_factors.shape[1]):  # Iterate over generative factors.
    for l in range(latents.shape[1]):
      ax = axes[l, j]
      if parents[l] != j:
        _visualize_interventional_effect(
            gen_factors, latents, l, parents[l], j, ax=ax, plot_legend=False)
        ax.set_title("")
      else:
        _visualize_interventional_effect(
            gen_factors, latents, l, parents[l], j, no_conditioning=True, ax=ax)
        ax.set_title("Parent={}, IRS = {:1.2}".format(parents[l], scores[l]))

  fig.tight_layout()
  if not tf.gfile.IsDirectory(output_dir):
    tf.gfile.MakeDirs(output_dir)
  output_path = os.path.join(output_dir, "interventional_effect.png")
  with tf.gfile.Open(output_path, "wb") as path:
    fig.savefig(path)


def _visualize_interventional_effect(gen_factors,
                                     latents,
                                     latent_dim,
                                     const_factor_idx,
                                     intervened_factor_idx,
                                     no_conditioning=False,
                                     ax=None,
                                     plot_legend=True,
                                     plot_scatter=False):
  """Visualize single cell of interventional effects.

  Args:
    gen_factors: Ground truth generative factors.
    latents: Latent factors.
    latent_dim: Latent dimension under consideration.
    const_factor_idx: Generative factor which is being kept constant.
    intervened_factor_idx: Generative factor on which we intervene.
    no_conditioning: Whether or not we should condition on const_factor_idx.
    ax: Matplotlib axis to use.
    plot_legend: Whether to plot a legend.
    plot_scatter: Whether to plot all points in a scatter plot.
  """
  if ax is None:
    plt.figure(figsize=(10, 7))
    ax = plt.axes()

  g_is = np.unique(gen_factors[:, const_factor_idx], axis=0)
  g_js = np.unique(gen_factors[:, intervened_factor_idx], axis=0)

  colors = ["b", "y", "g", "r", "c", "m", "k"]
  cols_idx = np.empty([gen_factors.shape[0]], dtype=int)
  for i_idx in range(g_is.shape[0]):
    match = (gen_factors[:, [const_factor_idx]] == [g_is[i_idx]]).all(axis=1)
    cols_idx[match] = i_idx
  cols = [colors[col % len(colors)] for col in cols_idx]

  # Plot all points, color indicates constant factor.
  if plot_scatter:
    ax.scatter(
        gen_factors[:, intervened_factor_idx], latents[:, latent_dim], c=cols)

  # Compute possible g_i and g_j.
  if no_conditioning:
    e_for_j = np.empty([g_js.shape[0]])
    median_for_j = np.empty([g_js.shape[0]])
    stdev_for_j = np.empty([g_js.shape[0]])
    for j_idx in range(g_js.shape[0]):
      match = (gen_factors[:, intervened_factor_idx] == g_js[j_idx])
      e_for_j[j_idx] = np.mean(latents[match, latent_dim])
      median_for_j[j_idx] = np.median(latents[match, latent_dim])
      stdev_for_j[j_idx] = np.std(latents[match, latent_dim])
    ax.plot(g_js, e_for_j, linewidth=2, markersize=12, label="mean")
    ax.plot(g_js, median_for_j, linewidth=2, markersize=12, label="median")
    ax.plot(g_js, e_for_j + stdev_for_j, linestyle="--", c="b", linewidth=1)
    ax.plot(g_js, e_for_j - stdev_for_j, linestyle="--", c="b", linewidth=1)
    ax.set_ylabel("E[z_{}|g_{}]".format(latent_dim, intervened_factor_idx))
    ax.set_xlabel("g_{}".format(intervened_factor_idx))
    ax.legend()
  else:
    # Compute E[Z_l | g_i, g_j] as a function of g_j for each g_i.
    e_given_i_for_j = np.empty([g_is.shape[0], g_js.shape[0]])
    for i_idx in range(g_is.shape[0]):
      for j_idx in range(g_js.shape[0]):
        match = (gen_factors[:, [const_factor_idx, intervened_factor_idx]] == [
            g_is[i_idx], g_js[j_idx]
        ]).all(axis=1)
        e_given_i_for_j[i_idx, j_idx] = np.mean(latents[match, latent_dim])
      ax.plot(
          g_js,
          e_given_i_for_j[i_idx, :],
          "go--",
          c=colors[i_idx % len(colors)],
          label="g_{}=={}".format(const_factor_idx, g_is[i_idx]),
          linewidth=1.5,
          markersize=3)

    ax.set_xlabel("int. g_{}".format(intervened_factor_idx))
    ax.set_ylabel("E[z_{}|g_{}, g_{}]".format(latent_dim, const_factor_idx,
                                              intervened_factor_idx))
    ax.set_title("Interventional Effect (keeping parent fixed)")
    if plot_legend:
      ax.legend()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization module for disentangled representations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numbers
import os
from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.utils import results
from disentanglement_lib.visualize import visualize_util
from disentanglement_lib.visualize.visualize_irs import vis_all_interventional_effects
import numpy as np
from scipy import stats
from six.moves import range
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import gfile
import tensorflow_hub as hub
import gin.tf


def visualize(model_dir,
              output_dir,
              overwrite=False,
              num_animations=5,
              num_frames=20,
              fps=10,
              num_points_irs=10000):
  """Takes trained model from model_dir and visualizes it in output_dir.

  Args:
    model_dir: Path to directory where the trained model is saved.
    output_dir: Path to output directory.
    overwrite: Boolean indicating whether to overwrite output directory.
    num_animations: Integer with number of distinct animations to create.
    num_frames: Integer with number of frames in each animation.
    fps: Integer with frame rate for the animation.
    num_points_irs: Number of points to be used for the IRS plots.
  """
  # Fix the random seed for reproducibility.
  random_state = np.random.RandomState(0)

  # Create the output directory if necessary.
  if tf.gfile.IsDirectory(output_dir):
    if overwrite:
      tf.gfile.DeleteRecursively(output_dir)
    else:
      raise ValueError("Directory already exists and overwrite is False.")

  # Automatically set the proper data set if necessary. We replace the active
  # gin config as this will lead to a valid gin config file where the data set
  # is present.
  # Obtain the dataset name from the gin config of the previous step.
  gin_config_file = os.path.join(model_dir, "results", "gin", "train.gin")
  gin_dict = results.gin_dict(gin_config_file)
  gin.bind_parameter("dataset.name", gin_dict["dataset.name"].replace(
      "'", ""))

  # Automatically infer the activation function from gin config.
  activation_str = gin_dict["reconstruction_loss.activation"]
  if activation_str == "'logits'":
    activation = sigmoid
  elif activation_str == "'tanh'":
    activation = tanh
  else:
    raise ValueError(
        "Activation function  could not be infered from gin config.")

  dataset = named_data.get_named_ground_truth_data()
  num_pics = 64
  module_path = os.path.join(model_dir, "tfhub")

  with hub.eval_function_for_module(module_path) as f:
    # Save reconstructions.
    real_pics = dataset.sample_observations(num_pics, random_state)
    raw_pics = f(
        dict(images=real_pics), signature="reconstructions",
        as_dict=True)["images"]
    pics = activation(raw_pics)
    paired_pics = np.concatenate((real_pics, pics), axis=2)
    paired_pics = [paired_pics[i, :, :, :] for i in range(paired_pics.shape[0])]
    results_dir = os.path.join(output_dir, "reconstructions")
    if not gfile.IsDirectory(results_dir):
      gfile.MakeDirs(results_dir)
    visualize_util.grid_save_images(
        paired_pics, os.path.join(results_dir, "reconstructions.jpg"))

    # Save samples.
    def _decoder(latent_vectors):
      return f(
          dict(latent_vectors=latent_vectors),
          signature="decoder",
          as_dict=True)["images"]

    num_latent = int(gin_dict["encoder.num_latent"])
    num_pics = 64
    random_codes = random_state.normal(0, 1, [num_pics, num_latent])
    pics = activation(_decoder(random_codes))
    results_dir = os.path.join(output_dir, "sampled")
    if not gfile.IsDirectory(results_dir):
      gfile.MakeDirs(results_dir)
    visualize_util.grid_save_images(pics,
                                    os.path.join(results_dir, "samples.jpg"))

    # Save latent traversals.
    result = f(
        dict(images=dataset.sample_observations(num_pics, random_state)),
        signature="gaussian_encoder",
        as_dict=True)
    means = result["mean"]
    logvars = result["logvar"]
    results_dir = os.path.join(output_dir, "traversals")
    if not gfile.IsDirectory(results_dir):
      gfile.MakeDirs(results_dir)
    for i in range(means.shape[1]):
      pics = activation(
          latent_traversal_1d_multi_dim(_decoder, means[i, :], None))
      file_name = os.path.join(results_dir, "traversals{}.jpg".format(i))
      visualize_util.grid_save_images([pics], file_name)

    # Save the latent traversal animations.
    results_dir = os.path.join(output_dir, "animated_traversals")
    if not gfile.IsDirectory(results_dir):
      gfile.MakeDirs(results_dir)

    # Cycle through quantiles of a standard Gaussian.
    for i, base_code in enumerate(means[:num_animations]):
      images = []
      for j in range(base_code.shape[0]):
        code = np.repeat(np.expand_dims(base_code, 0), num_frames, axis=0)
        code[:, j] = visualize_util.cycle_gaussian(base_code[j], num_frames)
        images.append(np.array(activation(_decoder(code))))
      filename = os.path.join(results_dir, "std_gaussian_cycle%d.gif" % i)
      visualize_util.save_animation(np.array(images), filename, fps)

    # Cycle through quantiles of a fitted Gaussian.
    for i, base_code in enumerate(means[:num_animations]):
      images = []
      for j in range(base_code.shape[0]):
        code = np.repeat(np.expand_dims(base_code, 0), num_frames, axis=0)
        loc = np.mean(means[:, j])
        total_variance = np.mean(np.exp(logvars[:, j])) + np.var(means[:, j])
        code[:, j] = visualize_util.cycle_gaussian(
            base_code[j], num_frames, loc=loc, scale=np.sqrt(total_variance))
        images.append(np.array(activation(_decoder(code))))
      filename = os.path.join(results_dir, "fitted_gaussian_cycle%d.gif" % i)
      visualize_util.save_animation(np.array(images), filename, fps)

    # Cycle through [-2, 2] interval.
    for i, base_code in enumerate(means[:num_animations]):
      images = []
      for j in range(base_code.shape[0]):
        code = np.repeat(np.expand_dims(base_code, 0), num_frames, axis=0)
        code[:, j] = visualize_util.cycle_interval(base_code[j], num_frames,
                                                   -2., 2.)
        images.append(np.array(activation(_decoder(code))))
      filename = os.path.join(results_dir, "fixed_interval_cycle%d.gif" % i)
      visualize_util.save_animation(np.array(images), filename, fps)

    # Cycle linearly through +-2 std dev of a fitted Gaussian.
    for i, base_code in enumerate(means[:num_animations]):
      images = []
      for j in range(base_code.shape[0]):
        code = np.repeat(np.expand_dims(base_code, 0), num_frames, axis=0)
        loc = np.mean(means[:, j])
        total_variance = np.mean(np.exp(logvars[:, j])) + np.var(means[:, j])
        scale = np.sqrt(total_variance)
        code[:, j] = visualize_util.cycle_interval(base_code[j], num_frames,
                                                   loc-2.*scale, loc+2.*scale)
        images.append(np.array(activation(_decoder(code))))
      filename = os.path.join(results_dir, "conf_interval_cycle%d.gif" % i)
      visualize_util.save_animation(np.array(images), filename, fps)

    # Cycle linearly through minmax of a fitted Gaussian.
    for i, base_code in enumerate(means[:num_animations]):
      images = []
      for j in range(base_code.shape[0]):
        code = np.repeat(np.expand_dims(base_code, 0), num_frames, axis=0)
        code[:, j] = visualize_util.cycle_interval(base_code[j], num_frames,
                                                   np.min(means[:, j]),
                                                   np.max(means[:, j]))
        images.append(np.array(activation(_decoder(code))))
      filename = os.path.join(results_dir, "minmax_interval_cycle%d.gif" % i)
      visualize_util.save_animation(np.array(images), filename, fps)

    # Interventional effects visualization.
    factors = dataset.sample_factors(num_points_irs, random_state)
    obs = dataset.sample_observations_from_factors(factors, random_state)
    latents = f(
        dict(images=obs), signature="gaussian_encoder", as_dict=True)["mean"]
    results_dir = os.path.join(output_dir, "interventional_effects")
    vis_all_interventional_effects(factors, latents, results_dir)

  # Finally, we clear the gin config that we have set.
  gin.clear_config()


def latent_traversal_1d_multi_dim(generator_fn,
                                  latent_vector,
                                  dimensions=None,
                                  values=None,
                                  transpose=False):
  """Creates latent traversals for a latent vector along multiple dimensions.

  Creates a 2d grid image where each grid image is generated by passing a
  modified version of latent_vector to the generator_fn. In each column, a
  fixed dimension of latent_vector is modified. In each row, the value in the
  modified dimension is replaced by a fixed value.

  Args:
    generator_fn: Function that computes (fixed size) images from latent
      representation. It should accept a single Numpy array argument of the same
      shape as latent_vector and return a Numpy array of images where the first
      dimension corresponds to the different vectors in latent_vectors.
    latent_vector: 1d Numpy array with the base latent vector to be used.
    dimensions: 1d Numpy array with the indices of the dimensions that should be
      modified. If an integer is passed, the dimensions 0, 1, ...,
      (dimensions - 1) are modified. If None is passed, all dimensions of
      latent_vector are modified.
    values: 1d Numpy array with the latent space values that should be used for
      modifications. If an integer is passed, a linear grid between -1 and 1
      with that many points is constructed. If None is passed, a default grid is
      used (whose specific design is not guaranteed).
    transpose: Boolean which indicates whether rows and columns of the 2d grid
      should be transposed.

  Returns:
    Numpy array with image.
  """
  if latent_vector.ndim != 1:
    raise ValueError("Latent vector needs to be 1-dimensional.")

  if dimensions is None:
    # Default case, use all available dimensions.
    dimensions = np.arange(latent_vector.shape[0])
  elif isinstance(dimensions, numbers.Integral):
    # Check that there are enough dimensions in latent_vector.
    if dimensions > latent_vector.shape[0]:
      raise ValueError("The number of dimensions of latent_vector is less than"
                       " the number of dimensions requested in the arguments.")
    if dimensions < 1:
      raise ValueError("The number of dimensions has to be at least 1.")
    dimensions = np.arange(dimensions)
  if dimensions.ndim != 1:
    raise ValueError("Dimensions vector needs to be 1-dimensional.")

  if values is None:
    # Default grid of values.
    values = np.linspace(-1., 1., num=11)
  elif isinstance(values, numbers.Integral):
    if values <= 1:
      raise ValueError("If an int is passed for values, it has to be >1.")
    values = np.linspace(-1., 1., num=values)
  if values.ndim != 1:
    raise ValueError("Values vector needs to be 1-dimensional.")

  # We iteratively generate the rows/columns for each dimension as different
  # Numpy arrays. We do not preallocate a single final Numpy array as this code
  # is not performance critical and as it reduces code complexity.
  num_values = len(values)
  row_or_columns = []
  for dimension in dimensions:
    # Creates num_values copy of the latent_vector along the first axis.
    latent_traversal_vectors = np.tile(latent_vector, [num_values, 1])
    # Intervenes in the latent space.
    latent_traversal_vectors[:, dimension] = values
    # Generate the batch of images
    images = generator_fn(latent_traversal_vectors)
    # Adds images as a row or column depending whether transpose is True.
    axis = (1 if transpose else 0)
    row_or_columns.append(np.concatenate(images, axis))
  axis = (0 if transpose else 1)
  return np.concatenate(row_or_columns, axis)


def sigmoid(x):
  return stats.logistic.cdf(x)


def tanh(x):
  return np.tanh(x) / 2. + .5



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions that are useful for the different metrics."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from six.moves import range
import sklearn
from sklearn import ensemble
from sklearn import linear_model
from sklearn import model_selection
import gin.tf


def generate_batch_factor_code(ground_truth_data, representation_function,
                               num_points, random_state, batch_size):
  """Sample a single training sample based on a mini-batch of ground-truth data.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observation as input and
      outputs a representation.
    num_points: Number of points to sample.
    random_state: Numpy random state used for randomness.
    batch_size: Batchsize to sample points.

  Returns:
    representations: Codes (num_codes, num_points)-np array.
    factors: Factors generating the codes (num_factors, num_points)-np array.
  """
  representations = None
  factors = None
  i = 0
  while i < num_points:
    num_points_iter = min(num_points - i, batch_size)
    current_factors, current_observations = \
        ground_truth_data.sample(num_points_iter, random_state)
    if i == 0:
      factors = current_factors
      representations = representation_function(current_observations)
    else:
      factors = np.vstack((factors, current_factors))
      representations = np.vstack((representations,
                                   representation_function(
                                       current_observations)))
    i += num_points_iter
  return np.transpose(representations), np.transpose(factors)


def split_train_test(observations, train_percentage):
  """Splits observations into a train and test set.

  Args:
    observations: Observations to split in train and test. They can be the
      representation or the observed factors of variation. The shape is
      (num_dimensions, num_points) and the split is over the points.
    train_percentage: Fraction of observations to be used for training.

  Returns:
    observations_train: Observations to be used for training.
    observations_test: Observations to be used for testing.
  """
  num_labelled_samples = observations.shape[1]
  num_labelled_samples_train = int(
      np.ceil(num_labelled_samples * train_percentage))
  num_labelled_samples_test = num_labelled_samples - num_labelled_samples_train
  observations_train = observations[:, :num_labelled_samples_train]
  observations_test = observations[:, num_labelled_samples_train:]
  assert observations_test.shape[1] == num_labelled_samples_test, \
      "Wrong size of the test set."
  return observations_train, observations_test


def obtain_representation(observations, representation_function, batch_size):
  """"Obtain representations from observations.

  Args:
    observations: Observations for which we compute the representation.
    representation_function: Function that takes observation as input and
      outputs a representation.
    batch_size: Batch size to compute the representation.
  Returns:
    representations: Codes (num_codes, num_points)-Numpy array.
  """
  representations = None
  num_points = observations.shape[0]
  i = 0
  while i < num_points:
    num_points_iter = min(num_points - i, batch_size)
    current_observations = observations[i:i + num_points_iter]
    if i == 0:
      representations = representation_function(current_observations)
    else:
      representations = np.vstack((representations,
                                   representation_function(
                                       current_observations)))
    i += num_points_iter
  return np.transpose(representations)


def discrete_mutual_info(mus, ys):
  """Compute discrete mutual information."""
  num_codes = mus.shape[0]
  num_factors = ys.shape[0]
  m = np.zeros([num_codes, num_factors])
  for i in range(num_codes):
    for j in range(num_factors):
      m[i, j] = sklearn.metrics.mutual_info_score(ys[j, :], mus[i, :])
  return m


def discrete_entropy(ys):
  """Compute discrete mutual information."""
  num_factors = ys.shape[0]
  h = np.zeros(num_factors)
  for j in range(num_factors):
    h[j] = sklearn.metrics.mutual_info_score(ys[j, :], ys[j, :])
  return h


@gin.configurable(
    "discretizer", blacklist=["target"])
def make_discretizer(target, num_bins=gin.REQUIRED,
                     discretizer_fn=gin.REQUIRED):
  """Wrapper that creates discretizers."""
  return discretizer_fn(target, num_bins)


@gin.configurable("histogram_discretizer", blacklist=["target"])
def _histogram_discretize(target, num_bins=gin.REQUIRED):
  """Discretization based on histograms."""
  discretized = np.zeros_like(target)
  for i in range(target.shape[0]):
    discretized[i, :] = np.digitize(target[i, :], np.histogram(
        target[i, :], num_bins)[1][:-1])
  return discretized


def normalize_data(data, mean=None, stddev=None):
  if mean is None:
    mean = np.mean(data, axis=1)
  if stddev is None:
    stddev = np.std(data, axis=1)
  return (data - mean[:, np.newaxis]) / stddev[:, np.newaxis], mean, stddev


@gin.configurable("predictor")
def make_predictor_fn(predictor_fn=gin.REQUIRED):
  """Wrapper that creates classifiers."""
  return predictor_fn


@gin.configurable("logistic_regression_cv")
def logistic_regression_cv():
  """Logistic regression with 5 folds cross validation."""
  return linear_model.LogisticRegressionCV(Cs=10,
                                           cv=model_selection.KFold(n_splits=5))


@gin.configurable("gradient_boosting_classifier")
def gradient_boosting_classifier():
  """Default gradient boosting classifier."""
  return ensemble.GradientBoostingClassifier()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the disentanglement metric from the BetaVAE paper.

Based on "beta-VAE: Learning Basic Visual Concepts with a Constrained
Variational Framework" (https://openreview.net/forum?id=Sy2fzU9gl).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import logging
import numpy as np
from six.moves import range
from sklearn import linear_model
import gin.tf


@gin.configurable(
    "beta_vae_sklearn",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_beta_vae_sklearn(ground_truth_data,
                             representation_function,
                             random_state,
                             artifact_dir=None,
                             batch_size=gin.REQUIRED,
                             num_train=gin.REQUIRED,
                             num_eval=gin.REQUIRED):
  """Computes the BetaVAE disentanglement metric using scikit-learn.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    batch_size: Number of points to be used to compute the training_sample.
    num_train: Number of points used for training.
    num_eval: Number of points used for evaluation.

  Returns:
    Dictionary with scores:
      train_accuracy: Accuracy on training set.
      eval_accuracy: Accuracy on evaluation set.
  """
  del artifact_dir
  logging.info("Generating training set.")
  train_points, train_labels = _generate_training_batch(
      ground_truth_data, representation_function, batch_size, num_train,
      random_state)

  logging.info("Training sklearn model.")
  model = linear_model.LogisticRegression(random_state=random_state)
  model.fit(train_points, train_labels)

  logging.info("Evaluate training set accuracy.")
  train_accuracy = model.score(train_points, train_labels)
  train_accuracy = np.mean(model.predict(train_points) == train_labels)
  logging.info("Training set accuracy: %.2g", train_accuracy)

  logging.info("Generating evaluation set.")
  eval_points, eval_labels = _generate_training_batch(
      ground_truth_data, representation_function, batch_size, num_eval,
      random_state)

  logging.info("Evaluate evaluation set accuracy.")
  eval_accuracy = model.score(eval_points, eval_labels)
  logging.info("Evaluation set accuracy: %.2g", eval_accuracy)
  scores_dict = {}
  scores_dict["train_accuracy"] = train_accuracy
  scores_dict["eval_accuracy"] = eval_accuracy
  return scores_dict


def _generate_training_batch(ground_truth_data, representation_function,
                             batch_size, num_points, random_state):
  """Sample a set of training samples based on a batch of ground-truth data.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    batch_size: Number of points to be used to compute the training_sample.
    num_points: Number of points to be sampled for training set.
    random_state: Numpy random state used for randomness.

  Returns:
    points: (num_points, dim_representation)-sized numpy array with training set
      features.
    labels: (num_points)-sized numpy array with training set labels.
  """
  points = None  # Dimensionality depends on the representation function.
  labels = np.zeros(num_points, dtype=np.int64)
  for i in range(num_points):
    labels[i], feature_vector = _generate_training_sample(
        ground_truth_data, representation_function, batch_size, random_state)
    if points is None:
      points = np.zeros((num_points, feature_vector.shape[0]))
    points[i, :] = feature_vector
  return points, labels


def _generate_training_sample(ground_truth_data, representation_function,
                              batch_size, random_state):
  """Sample a single training sample based on a mini-batch of ground-truth data.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observation as input and
      outputs a representation.
    batch_size: Number of points to be used to compute the training_sample
    random_state: Numpy random state used for randomness.

  Returns:
    index: Index of coordinate to be used.
    feature_vector: Feature vector of training sample.
  """
  # Select random coordinate to keep fixed.
  index = random_state.randint(ground_truth_data.num_factors)
  # Sample two mini batches of latent variables.
  factors1 = ground_truth_data.sample_factors(batch_size, random_state)
  factors2 = ground_truth_data.sample_factors(batch_size, random_state)
  # Ensure sampled coordinate is the same across pairs of samples.
  factors2[:, index] = factors1[:, index]
  # Transform latent variables to observation space.
  observation1 = ground_truth_data.sample_observations_from_factors(
      factors1, random_state)
  observation2 = ground_truth_data.sample_observations_from_factors(
      factors2, random_state)
  # Compute representations based on the observations.
  representation1 = representation_function(observation1)
  representation2 = representation_function(observation2)
  # Compute the feature vector based on differences in representation.
  feature_vector = np.mean(np.abs(representation1 - representation2), axis=0)
  return index, feature_vector



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dummy data sets used for testing."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.data.ground_truth import ground_truth_data


class IdentityObservationsData(ground_truth_data.GroundTruthData):
  """Data set where dummy factors are also the observations."""

  @property
  def num_factors(self):
    return 10

  @property
  def observation_shape(self):
    return 10

  @property
  def factors_num_values(self):
    return [1] * 10

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return random_state.random_integers(10, size=(num, self.num_factors))

  def sample_observations_from_factors(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    return factors

  @property
  def factor_names(self):
    return ["Factor {}".format(i) for i in range(self.num_factors)]


class DummyData(ground_truth_data.GroundTruthData):
  """Dummy image data set of random noise used for testing."""

  @property
  def num_factors(self):
    return 10

  @property
  def factors_num_values(self):
    return [5] * 10

  @property
  def observation_shape(self):
    return [64, 64, 1]

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return random_state.randint(5, size=(num, self.num_factors))

  def sample_observations_from_factors(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    return random_state.random_sample(size=(factors.shape[0], 64, 64, 1))




# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the dendrogram.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from disentanglement_lib.visualize import dendrogram
import numpy as np


class DendrogramTest(parameterized.TestCase):

  @parameterized.parameters((np.zeros((5, 10), dtype=np.float32),))
  def test_dendrogram_plot(self, matrix):
    dendrogram.dendrogram_plot(
        matrix, self.create_tempdir().full_path, ["Factor"] * matrix.shape[0])

  @parameterized.parameters(
      (
          # Step by step simulation of the union-find algorithm on the event
          # matrix [[1, 3, 4], [1, 2, 3], [0, 3, 2], [0, 2, 1]], we're now
          # considering the first step.
          np.array([0, 1, 2, 3]),  # Roots for nodes in the graph.
          1,  # Merge node 1 ...
          3,  # With node 3
          0,  # Threshold value of the considered edge: this is the first edge.
          np.array([[0., 0., 0., 0.]]),  # Current dendrogram matrix.
          {0: 0, 1: 1},  # Dictionary factors of variation to cluster IDs.
          {0: 1, 1: 1},  # Size of each cluster.
          np.array([[1, 2], [3, 4]]),  # Original edge-weight matrix.
          1,  # Max cluster ID we have so far.
          0,  # How many events have been found.
          np.array([[0., 0., 0., 0.]]),  # A factor and a code are merged.
          {0: 0, 1: 1},
          {0: 1, 1: 1},
          1, 0.
      ),
      (
          # We're now considering the second step. Node 1 and 3 were merged in
          # the previous step.
          np.array([0, 1, 2, 1]),  # Roots for nodes in the graph.
          1,  # Merge node 1 ...
          2,  # With node 2
          1,  # Threshold value of the considered edge: this is the first edge.
          np.array([[0., 0., 0., 0.]]),  # Current dendrogram matrix.
          {0: 0, 1: 1},  # Dictionary factors of variation to cluster IDs.
          {0: 1, 1: 1},  # Size of each cluster.
          np.array([[1, 2], [3, 4]]),  # Original edge-weight matrix.
          1,  # Max cluster ID we have so far.
          0,  # How many events have been found.
          np.array([[0., 0., 0., 0.]]),  # A factor and a code are merged.
          {0: 0, 1: 1},
          {0: 1, 1: 1},
          1, 0.
      ),
      (
          # We're now considering the third step. Cluster 1 and node 2 were
          # merged in the previous step. Cluster 0 and 1 merged into cluster 2
          # which has now size of 2.
          np.array([0, 1, 1, 1]),  # Nodes in the graph.
          0,  # Merge node 0 ...
          3,  # With node 3
          2,  # Threshold value of the considered edge: this is the second edge.
          np.array([[0., 0., 0., 0.]]),  # Current dendrogram matrix.
          {0: 0, 1: 1},  # Dictionary factors of variation to cluster IDs.
          {0: 1, 1: 1},  # Size of each cluster.
          np.array([[1, 2], [3, 4]]),  # Original edge-weight matrix.
          1,  # Max cluster ID we have so far.
          0,  # How many events have been found.
          np.array([[0., 1., 2., 2.]]),  # Two clusters merged.
          {0: 2, 1: 2},  # Both factors are now mapped to the new cluster 2.
          {0: 1, 1: 1, 2: 2},  # Size of the new cluster is added.
          2,  # Our new max cluster id is 2.
          1  # We have had our first cluster merge.
      ),
      (
          # We're now considering the last step. The last code gets merged into
          # the now only cluster. Since cluster 0 and 1 were merged, 0 is the
          # root of cluster 2.
          np.array([0, 0, 1, 1]),  # Nodes in the graph. Smallest node is root.
          0,  # Merge node 0 ...
          2,  # With node 2
          3,  # Threshold value of the considered edge: this is the first edge.
          np.array([[0., 1., 2., 2.]]),  # Current dendrogram matrix.
          {0: 2, 1: 2},  # Dictionary factors of variation to cluster IDs.
          {0: 1, 1: 1, 2: 2},  # Size of each cluster.
          np.array([[1, 2], [3, 4]]),  # Original edge-weight matrix.
          2,  # Max cluster ID we have so far.
          1,  # How many events have been found.
          np.array([[0., 1., 2., 2.]]),  # A code is merged into a cluster.
          {0: 2, 1: 2},  # Both factors are now mapped to the new cluster 2.
          {0: 1, 1: 1, 2: 2},  # Size of the new cluster is added.
          2,  # Our new max cluster id is 2.
          1  # We have had one cluster merge.
      ))
  def test_union(
      self, nodes, idx, idy, val, z, cluster_id, size, matrix, n_clusters,
      idx_found, z_target, cluster_id_target, size_target, n_clusters_target,
      idx_found_target):

    z, cluster_id, size, n_clusters, idx_found = dendrogram._union(
        nodes, idx, idy, val, z, cluster_id, size, matrix, n_clusters,
        idx_found)
    self.assertEqual((z_target == z).all(), True)
    self.assertEqual((cluster_id_target == cluster_id), True)
    self.assertEqual((size_target == size), True)
    self.assertEqual((n_clusters_target == n_clusters), True)
    self.assertEqual((idx_found_target == idx_found), True)


if __name__ == "__main__":
  absltest.main()




# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reduced downstream classification task.

Test downstream performance after removing the k most predictive features for
each factor of variation.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.metrics import dci
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
import gin.tf


@gin.configurable(
    "reduced_downstream_task",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_reduced_downstream_task(ground_truth_data,
                                    representation_function,
                                    random_state,
                                    artifact_dir=None,
                                    num_factors_to_remove=gin.REQUIRED,
                                    num_train=gin.REQUIRED,
                                    num_test=gin.REQUIRED,
                                    batch_size=16):
  """Computes loss of a reduced downstream task.

  Measure the information leakage in each latent component after removing the
  k ("factors_to_remove") most informative features for the prediction task.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_factors_to_remove: Number of factors to remove from the latent
      representation.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    batch_size: Batch size for sampling.

  Returns:
    Dictionary with scores.
  """
  del artifact_dir
  scores = {}
  # Loop on different sizes of the training 'batch', as specified with gin.
  for train_size in num_train:
    size_string = str(train_size)
    mus_train, ys_train = utils.generate_batch_factor_code(
        ground_truth_data, representation_function, train_size, random_state,
        batch_size)
    mus_test, ys_test = utils.generate_batch_factor_code(
        ground_truth_data, representation_function, num_test, random_state,
        batch_size)
    # Create variables for aggregated scores.
    reduced_factor_train_scores = []
    other_factors_train_scores = []
    reduced_factor_test_scores = []
    other_factors_test_scores = []
    # Compute the reduced representation and test it for each factor of
    # variation.
    for factor_of_interest in range(ground_truth_data.num_factors):
      # Copy the training data and eliminate the k most informative factors.
      reduced_mus_train = mus_train.copy()
      reduced_mus_test = mus_test.copy()
      for _ in range(num_factors_to_remove):
        reduced_mus_train, reduced_mus_test =\
          compute_reduced_representation(reduced_mus_train, ys_train,
                                         reduced_mus_test, ys_test,
                                         factor_of_interest)
      predictor_model = utils.make_predictor_fn()
      train_acc, test_acc = compute_predictive_accuracy(
          np.transpose(reduced_mus_train), ys_train,
          np.transpose(reduced_mus_test), ys_test, predictor_model)
      # Save scores for reduced factor.
      scores[size_string +
             ":reduced_factor_{}:mean_train_accuracy_reduced_factor".format(
                 factor_of_interest)] = train_acc[factor_of_interest]
      scores[size_string +
             ":reduced_factor_{}:mean_test_accuracy_reduced_factor".format(
                 factor_of_interest)] = test_acc[factor_of_interest]
      reduced_factor_train_scores.append(train_acc[factor_of_interest])
      reduced_factor_test_scores.append(test_acc[factor_of_interest])

      # Save the scores (accuracies) in the score dictionary.
      local_other_factors_train_scores = []
      local_other_factors_test_scores = []
      for i in range(len(train_acc)):
        scores[size_string +
               ":reduced_factor_{}:mean_train_accuracy_factor_{}".format(
                   factor_of_interest, i)] = train_acc[i]
        scores[size_string +
               ":reduced_factor_{}:mean_test_accuracy_factor_{}".format(
                   factor_of_interest, i)] = test_acc[i]
        if i != factor_of_interest:
          local_other_factors_train_scores.append(train_acc[i])
          local_other_factors_test_scores.append(test_acc[i])
      # Save mean score for non-reduced factors.
      scores[size_string +
             ":reduced_factor_{}:mean_train_accuracy_non_reduced_factor".format(
                 factor_of_interest)] = np.mean(
                     local_other_factors_train_scores)
      scores[size_string +
             ":reduced_factor_{}:mean_test_accuracy_non_reduced_factor".format(
                 factor_of_interest)] = np.mean(local_other_factors_test_scores)
      other_factors_train_scores.append(
          np.mean(local_other_factors_train_scores))
      other_factors_test_scores.append(np.mean(local_other_factors_test_scores))

    # Compute the aggregate scores.
    scores[size_string + ":mean_train_accuracy_reduced_factor"] = np.mean(
        reduced_factor_train_scores)
    scores[size_string + ":mean_test_accuracy_reduced_factor"] = np.mean(
        reduced_factor_test_scores)
    scores[size_string + ":mean_train_accuracy_other_factors"] = np.mean(
        other_factors_train_scores)
    scores[size_string + ":mean_test_accuracy_other_factors"] = np.mean(
        other_factors_test_scores)
  return scores


@gin.configurable("reduced_representation")
def compute_reduced_representation(mus_train,
                                   ys_train,
                                   mus_test,
                                   ys_test,
                                   factor_of_interest,
                                   correlation_measure=gin.REQUIRED):
  """Computes a reduced representation of the data.

  The most informative factor with respect to the labels is deleted.

  Args:
    mus_train: latent means of the training batch.
    ys_train: labels of the training batch.
    mus_test: latent means of the test batch.
    ys_test: labels of the test batch.
    factor_of_interest: index of the factor of interest.
    correlation_measure: measure of correlation.

  Returns:
    Tuple with reduced representations for the training and test set.
  """
  importance_matrix = correlation_measure(mus_train, ys_train, mus_test,
                                          ys_test)
  factor_of_interest_importance = importance_matrix[:, factor_of_interest]
  factor_to_remove_index = np.argmax(factor_of_interest_importance)
  # Remove the factor of variation above from the representation
  reduced_representation_train = np.delete(
      mus_train.copy(), factor_to_remove_index, axis=0)
  reduced_representation_test = np.delete(
      mus_test.copy(), factor_to_remove_index, axis=0)
  return reduced_representation_train, reduced_representation_test


@gin.configurable(
    "factorwise_dci",
    blacklist=["mus_train", "ys_train", "mus_test", "ys_test"])
def compute_factorwise_dci(mus_train, ys_train, mus_test, ys_test):
  """Computes the DCI importance matrix of the attributes.

  Args:
    mus_train: latent means of the training batch.
    ys_train: labels of the training batch.
    mus_test: latent means of the test batch.
    ys_test: labels of the test batch.

  Returns:
    Matrix with importance scores.
  """
  importance_matrix, _, _ = dci.compute_importance_gbt(mus_train, ys_train,
                                                       mus_test, ys_test)
  assert importance_matrix.shape[0] == mus_train.shape[0]
  assert importance_matrix.shape[1] == ys_train.shape[0]
  return importance_matrix


def compute_predictive_accuracy(x_train, y_train, x_test, y_test, predictor_fn):
  """Computes average predictive accuracy for train and test set.

  Args:
    x_train: data x of the training batch.
    y_train: labels y of the training batch.
    x_test: data x of the test batch.
    y_test: labels y of the test batch.
    predictor_fn: function that is used to fit and predict the labels.

  Returns:
    Tuple with lists of training and test set accuracies.
  """
  num_factors = y_train.shape[0]
  train_acc = []
  test_acc = []
  # Loop on the generative factors to predict
  for i in range(num_factors):
    model = predictor_fn()
    model.fit(x_train, y_train[i, :])
    train_acc.append(np.mean(model.predict(x_train) == y_train[i, :]))
    test_acc.append(np.mean(model.predict(x_test) == y_test[i, :]))
  return train_acc, test_acc



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interventional Robustness Score.

Based on the paper https://arxiv.org/abs/1811.00007.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import logging
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
import gin.tf


@gin.configurable(
    "irs",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_irs(ground_truth_data,
                representation_function,
                random_state,
                artifact_dir=None,
                diff_quantile=0.99,
                num_train=gin.REQUIRED,
                batch_size=gin.REQUIRED):
  """Computes the Interventional Robustness Score.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    diff_quantile: Float value between 0 and 1 to decide what quantile of diffs
      to select (use 1.0 for the version in the paper).
    num_train: Number of points used for training.
    batch_size: Batch size for sampling.

  Returns:
    Dict with IRS and number of active dimensions.
  """
  del artifact_dir
  logging.info("Generating training set.")
  mus, ys = utils.generate_batch_factor_code(ground_truth_data,
                                             representation_function, num_train,
                                             random_state, batch_size)
  assert mus.shape[1] == num_train

  ys_discrete = utils.make_discretizer(ys)
  active_mus = _drop_constant_dims(mus)

  if not active_mus.any():
    irs_score = 0.0
  else:
    irs_score = scalable_disentanglement_score(ys_discrete.T, active_mus.T,
                                               diff_quantile)["avg_score"]

  score_dict = {}
  score_dict["IRS"] = irs_score
  score_dict["num_active_dims"] = np.sum(active_mus)
  return score_dict


def _drop_constant_dims(ys):
  """Returns a view of the matrix `ys` with dropped constant rows."""
  ys = np.asarray(ys)
  if ys.ndim != 2:
    raise ValueError("Expecting a matrix.")

  variances = ys.var(axis=1)
  active_mask = variances > 0.
  return ys[active_mask, :]


def scalable_disentanglement_score(gen_factors, latents, diff_quantile=0.99):
  """Computes IRS scores of a dataset.

  Assumes no noise in X and crossed generative factors (i.e. one sample per
  combination of gen_factors). Assumes each g_i is an equally probable
  realization of g_i and all g_i are independent.

  Args:
    gen_factors: Numpy array of shape (num samples, num generative factors),
      matrix of ground truth generative factors.
    latents: Numpy array of shape (num samples, num latent dimensions), matrix
      of latent variables.
    diff_quantile: Float value between 0 and 1 to decide what quantile of diffs
      to select (use 1.0 for the version in the paper).

  Returns:
    Dictionary with IRS scores.
  """
  num_gen = gen_factors.shape[1]
  num_lat = latents.shape[1]

  # Compute normalizer.
  max_deviations = np.max(np.abs(latents - latents.mean(axis=0)), axis=0)
  cum_deviations = np.zeros([num_lat, num_gen])
  for i in range(num_gen):
    unique_factors = np.unique(gen_factors[:, i], axis=0)
    assert unique_factors.ndim == 1
    num_distinct_factors = unique_factors.shape[0]
    for k in range(num_distinct_factors):
      # Compute E[Z | g_i].
      match = gen_factors[:, i] == unique_factors[k]
      e_loc = np.mean(latents[match, :], axis=0)

      # Difference of each value within that group of constant g_i to its mean.
      diffs = np.abs(latents[match, :] - e_loc)
      max_diffs = np.percentile(diffs, q=diff_quantile*100, axis=0)
      cum_deviations[:, i] += max_diffs
    cum_deviations[:, i] /= num_distinct_factors
  # Normalize value of each latent dimension with its maximal deviation.
  normalized_deviations = cum_deviations / max_deviations[:, np.newaxis]
  irs_matrix = 1.0 - normalized_deviations
  disentanglement_scores = irs_matrix.max(axis=1)
  if np.sum(max_deviations) > 0.0:
    avg_score = np.average(disentanglement_scores, weights=max_deviations)
  else:
    avg_score = np.mean(disentanglement_scores)

  parents = irs_matrix.argmax(axis=1)
  score_dict = {}
  score_dict["disentanglement_scores"] = disentanglement_scores
  score_dict["avg_score"] = avg_score
  score_dict["parents"] = parents
  score_dict["IRS_matrix"] = irs_matrix
  score_dict["max_deviations"] = max_deviations
  return score_dict



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various utilities used in the data set code."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
from six.moves import range
import tensorflow.compat.v1 as tf


def tf_data_set_from_ground_truth_data(ground_truth_data, random_seed):
  """Generate a tf.data.DataSet from ground_truth data."""

  def generator():
    # We need to hard code the random seed so that the data set can be reset.
    random_state = np.random.RandomState(random_seed)
    while True:
      yield ground_truth_data.sample_observations(1, random_state)[0]

  return tf.data.Dataset.from_generator(
      generator, tf.float32, output_shapes=ground_truth_data.observation_shape)


class SplitDiscreteStateSpace(object):
  """State space with factors split between latent variable and observations."""

  def __init__(self, factor_sizes, latent_factor_indices):
    self.factor_sizes = factor_sizes
    self.num_factors = len(self.factor_sizes)
    self.latent_factor_indices = latent_factor_indices
    self.observation_factor_indices = [
        i for i in range(self.num_factors)
        if i not in self.latent_factor_indices
    ]

  @property
  def num_latent_factors(self):
    return len(self.latent_factor_indices)

  def sample_latent_factors(self, num, random_state):
    """Sample a batch of the latent factors."""
    factors = np.zeros(
        shape=(num, len(self.latent_factor_indices)), dtype=np.int64)
    for pos, i in enumerate(self.latent_factor_indices):
      factors[:, pos] = self._sample_factor(i, num, random_state)
    return factors

  def sample_all_factors(self, latent_factors, random_state):
    """Samples the remaining factors based on the latent factors."""
    num_samples = latent_factors.shape[0]
    all_factors = np.zeros(
        shape=(num_samples, self.num_factors), dtype=np.int64)
    all_factors[:, self.latent_factor_indices] = latent_factors
    # Complete all the other factors
    for i in self.observation_factor_indices:
      all_factors[:, i] = self._sample_factor(i, num_samples, random_state)
    return all_factors

  def _sample_factor(self, i, num, random_state):
    return random_state.randint(self.factor_sizes[i], size=num)


class StateSpaceAtomIndex(object):
  """Index mapping from features to positions of state space atoms."""

  def __init__(self, factor_sizes, features):
    """Creates the StateSpaceAtomIndex.

    Args:
      factor_sizes: List of integers with the number of distinct values for each
        of the factors.
      features: Numpy matrix where each row contains a different factor
        configuration. The matrix needs to cover the whole state space.
    """
    self.factor_sizes = factor_sizes
    num_total_atoms = np.prod(self.factor_sizes)
    self.factor_bases = num_total_atoms / np.cumprod(self.factor_sizes)
    feature_state_space_index = self._features_to_state_space_index(features)
    if np.unique(feature_state_space_index).size != num_total_atoms:
      raise ValueError("Features matrix does not cover the whole state space.")
    lookup_table = np.zeros(num_total_atoms, dtype=np.int64)
    lookup_table[feature_state_space_index] = np.arange(num_total_atoms)
    self.state_space_to_save_space_index = lookup_table

  def features_to_index(self, features):
    """Returns the indices in the input space for given factor configurations.

    Args:
      features: Numpy matrix where each row contains a different factor
        configuration for which the indices in the input space should be
        returned.
    """
    state_space_index = self._features_to_state_space_index(features)
    return self.state_space_to_save_space_index[state_space_index]

  def _features_to_state_space_index(self, features):
    """Returns the indices in the atom space for given factor configurations.

    Args:
      features: Numpy matrix where each row contains a different factor
        configuration for which the indices in the atom space should be
        returned.
    """
    if (np.any(features > np.expand_dims(self.factor_sizes, 0)) or
        np.any(features < 0)):
      raise ValueError("Feature indices have to be within [0, factor_size-1]!")
    return np.array(np.dot(features, self.factor_bases), dtype=np.int64)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for evaluate.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from absl.testing import parameterized
from disentanglement_lib.evaluation import evaluate
from disentanglement_lib.methods.unsupervised import train
from disentanglement_lib.postprocessing import postprocess
from disentanglement_lib.utils import resources
import gin.tf


class EvaluateTest(parameterized.TestCase):

  def setUp(self):
    super(EvaluateTest, self).setUp()
    self.model_dir = self.create_tempdir(
        "model", cleanup=absltest.TempFileCleanup.OFF).full_path
    model_config = resources.get_file(
        "config/tests/methods/unsupervised/train_test.gin")
    train.train_with_gin(self.model_dir, True, [model_config])
    self.output_dir = self.create_tempdir(
        "output", cleanup=absltest.TempFileCleanup.OFF).full_path
    postprocess_config = resources.get_file(
        "config/tests/postprocessing/postprocess_test_configs/mean.gin")
    postprocess.postprocess_with_gin(self.model_dir, self.output_dir, True,
                                     [postprocess_config])

  @parameterized.parameters(
      list(
          resources.get_files_in_folder(
              "config/tests/evaluation/evaluate_test_configs")))
  def test_evaluate(self, gin_config):
    # We clear the gin config before running. Otherwise, if a prior test fails,
    # the gin config is locked and the current test fails.
    gin.clear_config()
    evaluate.evaluate_with_gin(self.output_dir,
                               self.create_tempdir().full_path, True,
                               [gin_config])


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main training protocol used for computing abstract reasoning scores.

This is the main pipeline for the reasoning step in the paper:
Are Disentangled Representations Helpful for Abstract Visual Reasoning?
Sjoerd van Steenkiste, Francesco Locatello, Juergen Schmidhuber, Olivier Bachem.
NeurIPS, 2019.
https://arxiv.org/abs/1905.12506
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import time
from disentanglement_lib.evaluation.abstract_reasoning import models  # pylint: disable=unused-import
from disentanglement_lib.evaluation.abstract_reasoning import pgm_data
from disentanglement_lib.utils import results
import numpy as np
import tensorflow.compat.v1 as tf
import gin.tf.external_configurables  # pylint: disable=unused-import
import gin.tf
from tensorflow.contrib import tpu as contrib_tpu


def reason_with_gin(input_dir,
                    output_dir,
                    overwrite=False,
                    gin_config_files=None,
                    gin_bindings=None):
  """Trains a model based on the provided gin configuration.

  This function will set the provided gin bindings, call the reason() function
  and clear the gin config. Please see reason() for required gin bindings.

  Args:
    input_dir: String with path to directory where the representation is saved.
    output_dir: String with the path where the evaluation should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    gin_config_files: List of gin config files to load.
    gin_bindings: List of gin bindings to use.
  """
  if gin_config_files is None:
    gin_config_files = []
  if gin_bindings is None:
    gin_bindings = []
  gin.parse_config_files_and_bindings(gin_config_files, gin_bindings)
  reason(input_dir, output_dir, overwrite)
  gin.clear_config()


@gin.configurable(
    "abstract_reasoning", blacklist=["input_dir", "output_dir", "overwrite"])
def reason(
    input_dir,
    output_dir,
    overwrite=False,
    model=gin.REQUIRED,
    num_iterations=gin.REQUIRED,
    training_steps_per_iteration=gin.REQUIRED,
    eval_steps_per_iteration=gin.REQUIRED,
    random_seed=gin.REQUIRED,
    batch_size=gin.REQUIRED,
    name="",
):
  """Trains the estimator and exports the snapshot and the gin config.

  The use of this function requires the gin binding 'dataset.name' to be
  specified if a model is trained from scratch as that determines the data set
  used for training.

  Args:
    input_dir: String with path to directory where the representation function
      is saved.
    output_dir: String with the path where the results should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    model: GaussianEncoderModel that should be trained and exported.
    num_iterations: Integer with number of training steps.
    training_steps_per_iteration: Integer with number of training steps per
      iteration.
    eval_steps_per_iteration: Integer with number of validationand test steps
      per iteration.
    random_seed: Integer with random seed used for training.
    batch_size: Integer with the batch size.
    name: Optional string with name of the model (can be used to name models).
  """
  # We do not use the variable 'name'. Instead, it can be used to name results
  # as it will be part of the saved gin config.
  del name

  # Delete the output directory if it already exists.
  if tf.gfile.IsDirectory(output_dir):
    if overwrite:
      tf.gfile.DeleteRecursively(output_dir)
    else:
      raise ValueError("Directory already exists and overwrite is False.")

  # Create a numpy random state. We will sample the random seeds for training
  # and evaluation from this.
  random_state = np.random.RandomState(random_seed)

  # Automatically set the proper data set if necessary. We replace the active
  # gin config as this will lead to a valid gin config file where the data set
  # is present.
  if gin.query_parameter("dataset.name") == "auto":
    if input_dir is None:
      raise ValueError("Cannot automatically infer data set for methods with"
                       " no prior model directory.")
    # Obtain the dataset name from the gin config of the previous step.
    gin_config_file = os.path.join(input_dir, "results", "gin",
                                   "postprocess.gin")
    gin_dict = results.gin_dict(gin_config_file)
    with gin.unlock_config():
      gin.bind_parameter("dataset.name",
                         gin_dict["dataset.name"].replace("'", ""))
  dataset = pgm_data.get_pgm_dataset()

  # Set the path to the TFHub embedding if we are training based on a
  # pre-trained embedding..
  if input_dir is not None:
    tfhub_dir = os.path.join(input_dir, "tfhub")
    with gin.unlock_config():
      gin.bind_parameter("HubEmbedding.hub_path", tfhub_dir)

  # We create a TPUEstimator based on the provided model. This is primarily so
  # that we could switch to TPU training in the future. For now, we train
  # locally on GPUs.
  run_config = contrib_tpu.RunConfig(
      tf_random_seed=random_seed,
      keep_checkpoint_max=1,
      tpu_config=contrib_tpu.TPUConfig(iterations_per_loop=500))
  tpu_estimator = contrib_tpu.TPUEstimator(
      use_tpu=False,
      model_fn=model.model_fn,
      model_dir=os.path.join(output_dir, "tf_checkpoint"),
      train_batch_size=batch_size,
      eval_batch_size=batch_size,
      config=run_config)

  # Set up time to keep track of elapsed time in results.
  experiment_timer = time.time()

  # Create a dictionary to keep track of all relevant information.
  results_dict_of_dicts = {}
  validation_scores = []
  all_dicts = []

  for i in range(num_iterations):
    steps_so_far = i * training_steps_per_iteration
    tf.logging.info("Training to %d steps.", steps_so_far)
    # Train the model for the specified steps.
    tpu_estimator.train(
        input_fn=dataset.make_input_fn(random_state.randint(2**32)),
        steps=training_steps_per_iteration)
    # Compute validation scores used for model selection.
    validation_results = tpu_estimator.evaluate(
        input_fn=dataset.make_input_fn(
            random_state.randint(2**32), num_batches=eval_steps_per_iteration))
    validation_scores.append(validation_results["accuracy"])
    tf.logging.info("Validation results %s", validation_results)
    # Compute test scores for final results.
    test_results = tpu_estimator.evaluate(
        input_fn=dataset.make_input_fn(
            random_state.randint(2**32), num_batches=eval_steps_per_iteration),
        name="test")
    dict_at_iteration = results.namespaced_dict(
        val=validation_results, test=test_results)
    results_dict_of_dicts["step{}".format(steps_so_far)] = dict_at_iteration
    all_dicts.append(dict_at_iteration)

  # Select the best number of steps based on the validation scores and add it as
  # as a special key to the dictionary.
  best_index = np.argmax(validation_scores)
  results_dict_of_dicts["best"] = all_dicts[best_index]

  # Save the results. The result dir will contain all the results and config
  # files that we copied along, as we progress in the pipeline. The idea is that
  # these files will be available for analysis at the end.
  if input_dir is not None:
    original_results_dir = os.path.join(input_dir, "results")
  else:
    original_results_dir = None
  results_dict = results.namespaced_dict(**results_dict_of_dicts)
  results_dir = os.path.join(output_dir, "results")
  results_dict["elapsed_time"] = time.time() - experiment_timer
  results.update_result_directory(results_dir, "abstract_reasoning",
                                  results_dict, original_results_dir)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MPI3D data set."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from disentanglement_lib.data.ground_truth import ground_truth_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
import tensorflow.compat.v1 as tf



class MPI3D(ground_truth_data.GroundTruthData):
  """MPI3D dataset.

  MPI3D datasets have been introduced as a part of NEURIPS 2019 Disentanglement
  Competition.(http://www.disentanglement-challenge.com).
  There are three different datasets:
  1. Simplistic rendered images (mpi3d_toy).
  2. Realistic rendered images (mpi3d_realistic).
  3. Real world images (mpi3d_real).

  Currently only mpi3d_toy is publicly available. More details about this
  dataset can be found in "On the Transfer of Inductive Bias from Simulation to
  the Real World: a New Disentanglement Dataset"
  (https://arxiv.org/abs/1906.03292).

  The ground-truth factors of variation in the dataset are:
  0 - Object color (4 different values for the simulated datasets and 6 for the
    real one)
  1 - Object shape (4 different values for the simulated datasets and 6 for the
    real one)
  2 - Object size (2 different values)
  3 - Camera height (3 different values)
  4 - Background colors (3 different values)
  5 - First DOF (40 different values)
  6 - Second DOF (40 different values)
  """

  def __init__(self, mode="mpi3d_toy"):
    if mode == "mpi3d_toy":
      mpi3d_path = os.path.join(
          os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "mpi3d_toy",
          "mpi3d_toy.npz")
      if not tf.io.gfile.exists(mpi3d_path):
        raise ValueError(
            "Dataset '{}' not found. Make sure the dataset is publicly available and downloaded correctly."
            .format(mode))
      else:
        with tf.io.gfile.GFile(mpi3d_path, "rb") as f:
          data = np.load(f)
      self.factor_sizes = [4, 4, 2, 3, 3, 40, 40]
    elif mode == "mpi3d_realistic":
      mpi3d_path = os.path.join(
          os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "mpi3d_realistic",
          "mpi3d_realistic.npz")
      if not tf.io.gfile.exists(mpi3d_path):
        raise ValueError(
            "Dataset '{}' not found. Make sure the dataset is publicly available and downloaded correctly."
            .format(mode))
      else:
        with tf.io.gfile.GFile(mpi3d_path, "rb") as f:
          data = np.load(f)
      self.factor_sizes = [4, 4, 2, 3, 3, 40, 40]
    elif mode == "mpi3d_real":
      mpi3d_path = os.path.join(
          os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "mpi3d_real",
          "mpi3d_real.npz")
      if not tf.io.gfile.exists(mpi3d_path):
        raise ValueError(
            "Dataset '{}' not found. Make sure the dataset is publicly available and downloaded correctly."
            .format(mode))
      else:
        with tf.io.gfile.GFile(mpi3d_path, "rb") as f:
          data = np.load(f)
      self.factor_sizes = [6, 6, 2, 3, 3, 40, 40]
    else:
      raise ValueError("Unknown mode provided.")

    self.images = data["images"]
    self.latent_factor_indices = [0, 1, 2, 3, 4, 5, 6]
    self.num_total_factors = 7
    self.state_space = util.SplitDiscreteStateSpace(self.factor_sizes,
                                                    self.latent_factor_indices)
    self.factor_bases = np.prod(self.factor_sizes) / np.cumprod(
        self.factor_sizes)

  @property
  def num_factors(self):
    return self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return self.factor_sizes

  @property
  def observation_shape(self):
    return [64, 64, 3]


  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return self.state_space.sample_latent_factors(num, random_state)

  def sample_observations_from_factors(self, factors, random_state):
    all_factors = self.state_space.sample_all_factors(factors, random_state)
    indices = np.array(np.dot(all_factors, self.factor_bases), dtype=np.int64)
    return self.images[indices] / 255.



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the SAP score.

Based on "Variational Inference of Disentangled Latent Concepts from Unlabeled
Observations" (https://openreview.net/forum?id=H1kG7GZAW), Section 3.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import logging
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
from sklearn import svm
import gin.tf


@gin.configurable(
    "sap_score",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_sap(ground_truth_data,
                representation_function,
                random_state,
                artifact_dir=None,
                num_train=gin.REQUIRED,
                num_test=gin.REQUIRED,
                batch_size=16,
                continuous_factors=gin.REQUIRED):
  """Computes the SAP score.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing discrete variables.
    batch_size: Batch size for sampling.
    continuous_factors: Factors are continuous variable (True) or not (False).

  Returns:
    Dictionary with SAP score.
  """
  del artifact_dir
  logging.info("Generating training set.")
  mus, ys = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train,
      random_state, batch_size)
  mus_test, ys_test = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_test,
      random_state, batch_size)
  logging.info("Computing score matrix.")
  return _compute_sap(mus, ys, mus_test, ys_test, continuous_factors)


def _compute_sap(mus, ys, mus_test, ys_test, continuous_factors):
  """Computes score based on both training and testing codes and factors."""
  score_matrix = compute_score_matrix(mus, ys, mus_test,
                                      ys_test, continuous_factors)
  # Score matrix should have shape [num_latents, num_factors].
  assert score_matrix.shape[0] == mus.shape[0]
  assert score_matrix.shape[1] == ys.shape[0]
  scores_dict = {}
  scores_dict["SAP_score"] = compute_avg_diff_top_two(score_matrix)
  logging.info("SAP score: %.2g", scores_dict["SAP_score"])

  return scores_dict


@gin.configurable(
    "sap_score_validation",
    blacklist=["observations", "labels", "representation_function"])
def compute_sap_on_fixed_data(observations, labels, representation_function,
                              train_percentage=gin.REQUIRED,
                              continuous_factors=gin.REQUIRED,
                              batch_size=100):
  """Computes the SAP score on the fixed set of observations and labels.

  Args:
    observations: Observations on which to compute the score. Observations have
      shape (num_observations, 64, 64, num_channels).
    labels: Observed factors of variations.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    train_percentage: Percentage of observations used for training.
    continuous_factors: Whether factors should be considered continuous or
      discrete.
    batch_size: Batch size used to compute the representation.

  Returns:
    SAP computed on the provided observations and labels.
  """
  mus = utils.obtain_representation(observations, representation_function,
                                    batch_size)
  assert labels.shape[1] == observations.shape[0], "Wrong labels shape."
  assert mus.shape[1] == observations.shape[0], "Wrong representation shape."
  mus_train, mus_test = utils.split_train_test(
      mus,
      train_percentage)
  ys_train, ys_test = utils.split_train_test(
      labels,
      train_percentage)
  return _compute_sap(mus_train, ys_train, mus_test, ys_test,
                      continuous_factors)


def compute_score_matrix(mus, ys, mus_test, ys_test, continuous_factors):
  """Compute score matrix as described in Section 3."""
  num_latents = mus.shape[0]
  num_factors = ys.shape[0]
  score_matrix = np.zeros([num_latents, num_factors])
  for i in range(num_latents):
    for j in range(num_factors):
      mu_i = mus[i, :]
      y_j = ys[j, :]
      if continuous_factors:
        # Attribute is considered continuous.
        cov_mu_i_y_j = np.cov(mu_i, y_j, ddof=1)
        cov_mu_y = cov_mu_i_y_j[0, 1]**2
        var_mu = cov_mu_i_y_j[0, 0]
        var_y = cov_mu_i_y_j[1, 1]
        if var_mu > 1e-12:
          score_matrix[i, j] = cov_mu_y * 1. / (var_mu * var_y)
        else:
          score_matrix[i, j] = 0.
      else:
        # Attribute is considered discrete.
        mu_i_test = mus_test[i, :]
        y_j_test = ys_test[j, :]
        classifier = svm.LinearSVC(C=0.01, class_weight="balanced")
        classifier.fit(mu_i[:, np.newaxis], y_j)
        pred = classifier.predict(mu_i_test[:, np.newaxis])
        score_matrix[i, j] = np.mean(pred == y_j_test)
  return score_matrix


def compute_avg_diff_top_two(matrix):
  sorted_matrix = np.sort(matrix, axis=0)
  return np.mean(sorted_matrix[-1, :] - sorted_matrix[-2, :])



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Postprocessing step that extracts representation from trained model."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import time
from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.postprocessing import methods  # pylint: disable=unused-import
from disentanglement_lib.utils import convolute_hub
from disentanglement_lib.utils import results
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub

import gin.tf


def postprocess_with_gin(model_dir,
                         output_dir,
                         overwrite=False,
                         gin_config_files=None,
                         gin_bindings=None):
  """Postprocess a trained model based on the provided gin configuration.

  This function will set the provided gin bindings, call the postprocess()
  function and clear the gin config. Please see the postprocess() for required
  gin bindings.

  Args:
    model_dir: String with path to directory where the model is saved.
    output_dir: String with the path where the representation should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    gin_config_files: List of gin config files to load.
    gin_bindings: List of gin bindings to use.
  """
  if gin_config_files is None:
    gin_config_files = []
  if gin_bindings is None:
    gin_bindings = []
  gin.parse_config_files_and_bindings(gin_config_files, gin_bindings)
  postprocess(model_dir, output_dir, overwrite)
  gin.clear_config()


@gin.configurable(
    "postprocess", blacklist=["model_dir", "output_dir", "overwrite"])
def postprocess(model_dir,
                output_dir,
                overwrite=False,
                postprocess_fn=gin.REQUIRED,
                random_seed=gin.REQUIRED,
                name=""):
  """Loads a trained Gaussian encoder and extracts representation.

  Args:
    model_dir: String with path to directory where the model is saved.
    output_dir: String with the path where the representation should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    postprocess_fn: Function used to extract the representation (see methods.py
      for examples).
    random_seed: Integer with random seed used for postprocessing (may be
      unused).
    name: Optional string with name of the representation (can be used to name
      representations).
  """
  # We do not use the variable 'name'. Instead, it can be used to name
  # representations as it will be part of the saved gin config.
  del name

  # Delete the output directory if it already exists.
  if tf.gfile.IsDirectory(output_dir):
    if overwrite:
      tf.gfile.DeleteRecursively(output_dir)
    else:
      raise ValueError("Directory already exists and overwrite is False.")

  # Set up timer to keep track of elapsed time in results.
  experiment_timer = time.time()

  # Automatically set the proper data set if necessary. We replace the active
  # gin config as this will lead to a valid gin config file where the data set
  # is present.
  if gin.query_parameter("dataset.name") == "auto":
    # Obtain the dataset name from the gin config of the previous step.
    gin_config_file = os.path.join(model_dir, "results", "gin", "train.gin")
    gin_dict = results.gin_dict(gin_config_file)
    with gin.unlock_config():
      gin.bind_parameter("dataset.name", gin_dict["dataset.name"].replace(
          "'", ""))
  dataset = named_data.get_named_ground_truth_data()

  # Path to TFHub module of previously trained model.
  module_path = os.path.join(model_dir, "tfhub")
  with hub.eval_function_for_module(module_path) as f:

    def _gaussian_encoder(x):
      """Encodes images using trained model."""
      # Push images through the TFHub module.
      output = f(dict(images=x), signature="gaussian_encoder", as_dict=True)
      # Convert to numpy arrays and return.
      return {key: np.array(values) for key, values in output.items()}

    # Run the postprocessing function which returns a transformation function
    # that can be used to create the representation from the mean and log
    # variance of the Gaussian distribution given by the encoder. Also returns
    # path to a checkpoint if the transformation requires variables.
    transform_fn, transform_checkpoint_path = postprocess_fn(
        dataset, _gaussian_encoder, np.random.RandomState(random_seed),
        output_dir)

    # Takes the "gaussian_encoder" signature, extracts the representation and
    # then saves under the signature "representation".
    tfhub_module_dir = os.path.join(output_dir, "tfhub")
    convolute_hub.convolute_and_save(
        module_path, "gaussian_encoder", tfhub_module_dir, transform_fn,
        transform_checkpoint_path, "representation")

  # We first copy over all the prior results and configs.
  original_results_dir = os.path.join(model_dir, "results")
  results_dir = os.path.join(output_dir, "results")
  results_dict = dict(elapsed_time=time.time() - experiment_timer)
  results.update_result_directory(results_dir, "postprocess", results_dict,
                                  original_results_dir)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for vae.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import parameterized
from disentanglement_lib.methods.unsupervised import vae
import numpy as np
import tensorflow.compat.v1 as tf
import gin.tf.external_configurables  # pylint: disable=unused-import


def _make_symmetric_psd(matrix):
  return 0.5 * (matrix + matrix.T) + np.diag(np.ones(10)) * 10.


class VaeTest(parameterized.TestCase, tf.test.TestCase):

  @parameterized.parameters((np.zeros([10, 10]), np.zeros([10, 10]), 0., 0.01),
                            (np.ones([10, 10]), np.zeros([10, 10]), 5., 5.01),
                            (np.ones([10, 10]), np.ones([10, 10]), 8.58, 8.6))
  def test_compute_gaussian_kl(self, mean, logvar, target_low, target_high):
    mean_tf = tf.convert_to_tensor(mean, dtype=np.float32)
    logvar_tf = tf.convert_to_tensor(logvar, dtype=np.float32)
    with self.test_session() as sess:
      test_value = sess.run(vae.compute_gaussian_kl(mean_tf, logvar_tf))
      self.assertBetween(test_value, target_low, target_high)

  @parameterized.parameters((0, 0., 0.01), (10, 10., 10.01),
                            (100, 100., 100.01), (101, 100., 100.01))
  def test_anneal(self, step, target_low, target_high):
    c_max = 100.
    iteration_threshold = 100
    with self.test_session() as sess:
      test_value = sess.run(vae.anneal(c_max, step, iteration_threshold))
      self.assertBetween(test_value, target_low, target_high)

  @parameterized.parameters(
      (True, 0., 1.), (True, 0., 4.), (True, 1., 1.),
      (False, np.zeros(10), np.ones([10, 10])),
      (False, np.zeros(10), _make_symmetric_psd(np.random.random((10, 10)))))
  def test_compute_covariance_z_mean(self, diagonal, mean, cov):
    if diagonal:
      samples = tf.random.normal(
          shape=(100000, 10), mean=mean, stddev=tf.math.sqrt(cov))
      cov = np.diag(np.ones([10])) * cov
    else:
      samples = tf.constant(
          np.random.multivariate_normal(mean, cov, size=(1000000)))
    with self.test_session() as sess:
      test_value = sess.run(vae.compute_covariance_z_mean(samples))
      self.assertBetween(np.sum((test_value - cov)**2), 0., 0.1)

  @parameterized.parameters(
      (np.ones([10, 10]), 90., 90.1), (np.zeros([10, 10]), 10., 10.1),
      (np.diag(np.ones(10)), 0., 0.1), (2. * np.diag(np.ones(10)), 10., 10.1))
  def test_regularize_diag_off_diag_dip(self, matrix, target_low, target_high):
    matrix_tf = tf.convert_to_tensor(matrix, dtype=np.float32)
    with self.test_session() as sess:
      test_value = sess.run(vae.regularize_diag_off_diag_dip(matrix_tf, 1, 1))
      self.assertBetween(test_value, target_low, target_high)

  @parameterized.parameters((0., -1.4190, -1.4188), (1., -0.92, -0.91))
  def test_gaussian_log_density(self, z_mean, target_low, target_high):
    matrix = tf.ones(1)
    with self.test_session() as sess:
      test_value = sess.run(vae.gaussian_log_density(matrix, z_mean, 0.))[0]
      self.assertBetween(test_value, target_low, target_high)

  @parameterized.parameters(
      (1, 0., 0.1), (10, -82.9, -82.89))  # -82.893 = (10 - 1) * ln(10000)
  def test_total_correlation(self, num_dim, target_low, target_high):
    # Since there is no dataset, the constant should be (num_latent - 1)*log(N)
    z = tf.random.normal(shape=(10000, num_dim))
    z_mean = tf.zeros(shape=(10000, num_dim))
    z_logvar = tf.zeros(shape=(10000, num_dim))
    with self.test_session() as sess:
      test_value = sess.run(vae.total_correlation(z, z_mean, z_logvar))
      self.assertBetween(test_value, target_low, target_high)


if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for irs.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.evaluation.metrics import irs
import numpy as np
import gin.tf


def _identity_discretizer(target, num_bins):
  del num_bins
  return target


class IrsTest(absltest.TestCase):

  def test_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)

    ground_truth_data = dummy_data.IdentityObservationsData()
    representation_function = lambda x: np.array(x, dtype=np.float64)
    random_state = np.random.RandomState(0)
    scores = irs.compute_irs(ground_truth_data, representation_function,
                             random_state, None, 0.99, 3000, 3000)
    self.assertBetween(scores["IRS"], 0.9, 1.0)

  def test_bad_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)

    ground_truth_data = dummy_data.IdentityObservationsData()
    representation_function = lambda x: np.zeros_like(x, dtype=np.float64)
    random_state = np.random.RandomState(0)
    scores = irs.compute_irs(ground_truth_data, representation_function,
                             random_state, None, 0.99, 3000, 3000)
    self.assertBetween(scores["IRS"], 0.0, 0.1)

  def test_drop_constant_dims(self):
    random_state = np.random.RandomState(0)
    ys = random_state.normal(0.0, 1.0, (100, 100))
    ys[0, :] = 1.
    ys[-1, :] = 0.
    active_ys = irs._drop_constant_dims(ys)
    np.testing.assert_array_equal(active_ys, ys[1:-1])


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Downstream classification task."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
import gin.tf


@gin.configurable(
    "downstream_task",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_downstream_task(ground_truth_data,
                            representation_function,
                            random_state,
                            artifact_dir=None,
                            num_train=gin.REQUIRED,
                            num_test=gin.REQUIRED,
                            batch_size=16):
  """Computes loss of downstream task.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    batch_size: Batch size for sampling.

  Returns:
    Dictionary with scores.
  """
  del artifact_dir
  scores = {}
  for train_size in num_train:
    mus_train, ys_train = utils.generate_batch_factor_code(
        ground_truth_data, representation_function, train_size, random_state,
        batch_size)
    mus_test, ys_test = utils.generate_batch_factor_code(
        ground_truth_data, representation_function, num_test, random_state,
        batch_size)
    predictor_model = utils.make_predictor_fn()

    train_err, test_err = _compute_loss(
        np.transpose(mus_train), ys_train, np.transpose(mus_test),
        ys_test, predictor_model)
    size_string = str(train_size)
    scores[size_string +
           ":mean_train_accuracy"] = np.mean(train_err)
    scores[size_string +
           ":mean_test_accuracy"] = np.mean(test_err)
    scores[size_string +
           ":min_train_accuracy"] = np.min(train_err)
    scores[size_string + ":min_test_accuracy"] = np.min(test_err)
    for i in range(len(train_err)):
      scores[size_string +
             ":train_accuracy_factor_{}".format(i)] = train_err[i]
      scores[size_string + ":test_accuracy_factor_{}".format(i)] = test_err[i]

  return scores


def _compute_loss(x_train, y_train, x_test, y_test, predictor_fn):
  """Compute average accuracy for train and test set."""
  num_factors = y_train.shape[0]
  train_loss = []
  test_loss = []
  for i in range(num_factors):
    model = predictor_fn()
    model.fit(x_train, y_train[i, :])
    train_loss.append(np.mean(model.predict(x_train) == y_train[i, :]))
    test_loss.append(np.mean(model.predict(x_test) == y_test[i, :]))
  return train_loss, test_loss



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Downstream out-of-distribution classification task."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
from six.moves import range
import gin.tf


@gin.configurable(
    "strong_downstream_task",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_strong_downstream_task(ground_truth_data,
                                   representation_function,
                                   random_state,
                                   artifact_dir=None,
                                   num_train=gin.REQUIRED,
                                   num_test=gin.REQUIRED,
                                   n_experiment=gin.REQUIRED):
  """Computes loss of downstream task.

  This task is about strong generalization under covariate shifts. We first
  perform an intervention fixing a value for a factor in the whole training set.
  Then, we train a GBT classifier, and at test time, we consider all other
  values for that factor. We repeat the experiment n_experiment times, to ensure
  robustness.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    n_experiment: Number of repetitions of the experiment.

  Returns:
    Dictionary with scores.
  """
  del artifact_dir
  scores = {}
  for train_size in num_train:
    # sample factors
    factors_train = ground_truth_data.sample_factors(train_size, random_state)
    factors_test = ground_truth_data.sample_factors(num_test, random_state)
    # obtain_observations without intervention
    x_train = ground_truth_data.sample_observations_from_factors(
        factors_train, random_state)
    x_test = ground_truth_data.sample_observations_from_factors(
        factors_test, random_state)
    mus_train = representation_function(x_train)
    mus_test = representation_function(x_test)
    # train predictor on data without interbention
    predictor_model = utils.make_predictor_fn()
    y_train = np.transpose(factors_train)
    y_test = np.transpose(factors_test)
    train_err, test_err = _compute_loss(
        mus_train, y_train, mus_test,
        y_test, predictor_model)

    # train predictor on data with interventions
    train_err_int, test_err_int = _compute_loss_intervene(
        factors_train, factors_test, predictor_model, ground_truth_data,
        representation_function, random_state, n_experiment)

    size_string = str(train_size)
    scores[size_string +
           ":mean_train_accuracy"] = np.mean(train_err)
    scores[size_string +
           ":mean_test_accuracy"] = np.mean(test_err)
    scores[size_string +
           ":mean_strong_train_accuracy"] = np.mean(train_err_int)
    scores[size_string +
           ":mean_strong_test_accuracy"] = np.mean(test_err_int)
    scores[size_string + ":strong_generalization_gap"] = 1. - (
        scores[size_string + ":mean_strong_test_accuracy"] /
        scores[size_string + ":mean_test_accuracy"])
  return scores


def _compute_loss(x_train, y_train, x_test, y_test, predictor_fn):
  """Compute average accuracy for train and test set."""
  num_factors = y_train.shape[0]
  train_loss = []
  test_loss = []
  for i in range(num_factors):
    model = predictor_fn()
    model.fit(x_train, y_train[i, :])
    train_loss.append(np.mean(model.predict(x_train) == y_train[i, :]))
    test_loss.append(np.mean(model.predict(x_test) == y_test[i, :]))
  return train_loss, test_loss


def _compute_loss_intervene(factors_train, factors_test, predictor_fn,
                            ground_truth_data, representation_function,
                            random_state, n_experiment=10):
  """Compute average accuracy for train and test set."""
  num_factors = factors_train.shape[1]
  train_loss = []
  test_loss = []
  for i in range(num_factors):
    for _ in range(n_experiment):
      factors_train_int, factors_test_int, _, _ = intervene(
          factors_train.copy(), factors_test.copy(), i, num_factors,
          ground_truth_data)

      obs_train_int = ground_truth_data.sample_observations_from_factors(
          factors_train_int, random_state)
      obs_test_int = ground_truth_data.sample_observations_from_factors(
          factors_test_int, random_state)

      x_train_int = representation_function(obs_train_int)
      x_test_int = representation_function(obs_test_int)
      # train predictor on data without intervention
      y_train_int = np.transpose(factors_train_int)
      y_test_int = np.transpose(factors_test_int)
      model = predictor_fn()
      model.fit(x_train_int, y_train_int[i, :])
      train_loss.append(
          np.mean(model.predict(x_train_int) == y_train_int[i, :]))
      test_loss.append(np.mean(model.predict(x_test_int) == y_test_int[i, :]))
  return train_loss, test_loss


def intervene(y_train, y_test, target_y, num_factors, ground_truth_data):
  """Make random intervention on training data and remove it from test."""
  # sample coordinate to intervene on
  factor_list_to_interv = [j for j in range(num_factors) if j != target_y]
  # pick a factor to intervene on
  interv_factor = np.random.choice(factor_list_to_interv)
  # get the factor range
  all_val_factor = list(range(
      ground_truth_data.factors_num_values[interv_factor]))
  factor_interv_train = np.random.choice(all_val_factor)
  all_val_factor.remove(factor_interv_train)
  y_train[:, interv_factor] = factor_interv_train

  factor_interv_test = np.random.choice(all_val_factor,
                                        size=y_test.shape[0])
  y_test[:, interv_factor] = factor_interv_test
  return y_train, y_test, interv_factor, factor_interv_train



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for optimizer.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import parameterized
from disentanglement_lib.methods.shared import optimizers
from six.moves import range
import tensorflow.compat.v1 as tf

import gin.tf.external_configurables  # pylint: disable=unused-import
import gin.tf


def _make_vae_optimizer_configs():
  """Yield different vae_optimizer test configurations.

  Yields:
    A tuple containing a list of gin bindings, and the expected learning rate
    after 10 steps.
  """
  # Constant learning rate specified in the optimizer.
  bindings = [
      "vae_optimizer.optimizer_fn = @GradientDescentOptimizer",
      "GradientDescentOptimizer.learning_rate = 0.1",
  ]
  yield (bindings, 0.1)

  # Constant learning rate specified in vae_optimizer.
  bindings = [
      "vae_optimizer.optimizer_fn = @GradientDescentOptimizer",
      "vae_optimizer.learning_rate = 0.1",
  ]
  yield (bindings, 0.1)

  # Piecewise constant learning rate.
  bindings = [
      "vae_optimizer.optimizer_fn = @GradientDescentOptimizer",
      "vae_optimizer.learning_rate = @piecewise_constant",
      "piecewise_constant.boundaries = (3, 5)",
      "piecewise_constant.values = (0.2, 0.1, 0.01)",
  ]
  yield (bindings, 0.01)

  # Exponential decay learning rate.
  bindings = [
      "vae_optimizer.optimizer_fn = @GradientDescentOptimizer",
      "vae_optimizer.learning_rate = @exponential_decay",
      "exponential_decay.learning_rate = 0.1",
      "exponential_decay.decay_steps = 1",
      "exponential_decay.decay_rate = 0.9",
  ]
  yield (bindings, 0.03486784401)


class OptimizerTest(parameterized.TestCase, tf.test.TestCase):

  @parameterized.parameters(list(_make_vae_optimizer_configs()))
  def test_vae_optimizer(self, gin_bindings, expected_learning_rate):
    gin.parse_config_files_and_bindings([], gin_bindings)

    with self.test_session():
      x = tf.Variable(0.0)
      y = tf.pow(x + 2.0, 2.0)
      global_step = tf.train.get_or_create_global_step()
      optimizer = optimizers.make_vae_optimizer()
      train_op = optimizer.minimize(loss=y, global_step=global_step)
      tf.global_variables_initializer().run()
      for it in range(10):
        self.evaluate(train_op)
        self.assertEqual(it + 1, self.evaluate(global_step))
      current_learning_rate = self.evaluate(optimizer._learning_rate_tensor)
      self.assertAlmostEqual(expected_learning_rate, current_learning_rate)

    gin.clear_config()


if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utils.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.evaluation.metrics import utils
import numpy as np


class UtilsTest(absltest.TestCase):

  def test_histogram_discretizer(self):
    # Input of 2D samples.
    target = np.array([[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
                       [0.6, .5, .4, .3, .2, .1]])
    result = utils._histogram_discretize(target, num_bins=3)
    shouldbe = np.array([[1, 1, 2, 2, 3, 3], [3, 3, 2, 2, 1, 1]])
    np.testing.assert_array_equal(result, shouldbe)

  def test_discrete_entropy(self):
    target = np.array([[1, 1, 2, 2, 3, 3], [3, 3, 2, 2, 1, 1]])
    result = utils.discrete_entropy(target)
    shouldbe = np.log(3)
    np.testing.assert_allclose(result, [shouldbe, shouldbe])

  def test_discrete_mutual_info(self):
    xs = np.array([[1, 2, 1, 2], [1, 1, 2, 2]])
    ys = np.array([[1, 2, 1, 2], [2, 2, 1, 1]])
    result = utils.discrete_mutual_info(xs, ys)
    shouldbe = np.array([[np.log(2), 0.], [0., np.log(2)]])
    np.testing.assert_allclose(result, shouldbe)

  def test_split_train_test(self):
    xs = np.zeros([10, 100])
    xs_train, xs_test = utils.split_train_test(xs, 0.9)
    shouldbe_train = np.zeros([10, 90])
    shouldbe_test = np.zeros([10, 10])
    np.testing.assert_allclose(xs_train, shouldbe_train)
    np.testing.assert_allclose(xs_test, shouldbe_test)

if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods to visualize latent factors in the data sets."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.visualize import visualize_util
import numpy as np
from six.moves import range
from tensorflow.compat.v1 import gfile


def visualize_dataset(dataset_name, output_path, num_animations=5,
                      num_frames=20, fps=10):
  """Visualizes the data set by saving images to output_path.

  For each latent factor, outputs 16 images where only that latent factor is
  varied while all others are kept constant.

  Args:
    dataset_name: String with name of dataset as defined in named_data.py.
    output_path: String with path in which to create the visualizations.
    num_animations: Integer with number of distinct animations to create.
    num_frames: Integer with number of frames in each animation.
    fps: Integer with frame rate for the animation.
  """
  data = named_data.get_named_ground_truth_data(dataset_name)
  random_state = np.random.RandomState(0)

  # Create output folder if necessary.
  path = os.path.join(output_path, dataset_name)
  if not gfile.IsDirectory(path):
    gfile.MakeDirs(path)

  # Create still images.
  for i in range(data.num_factors):
    factors = data.sample_factors(16, random_state)
    indices = [j for j in range(data.num_factors) if i != j]
    factors[:, indices] = factors[0, indices]
    images = data.sample_observations_from_factors(factors, random_state)
    visualize_util.grid_save_images(
        images, os.path.join(path, "variations_of_factor%s.png" % i))

  # Create animations.
  for i in range(num_animations):
    base_factor = data.sample_factors(1, random_state)
    images = []
    for j, num_atoms in enumerate(data.factors_num_values):
      factors = np.repeat(base_factor, num_frames, axis=0)
      factors[:, j] = visualize_util.cycle_factor(base_factor[0, j], num_atoms,
                                                  num_frames)
      images.append(data.sample_observations_from_factors(factors,
                                                          random_state))
    visualize_util.save_animation(np.array(images),
                                  os.path.join(path, "animation%d.gif" % i),
                                  fps)




# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for visualize_dataset.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.visualize import visualize_dataset


class VisualizeDatasetTest(absltest.TestCase):

  def test_visualize(self):
    visualize_dataset.visualize_dataset("dummy_data",
                                        self.create_tempdir().full_path)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for the visualization code."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import math
from disentanglement_lib.utils import resources
import numpy as np
from PIL import Image
import scipy
from six.moves import range
import tensorflow.compat.v1 as tf
import imageio


def save_image(image, image_path):
  """Saves an image in the [0,1]-valued Numpy array to image_path.

  Args:
    image: Numpy array of shape (height, width, {1,3}) with values in [0, 1].
    image_path: String with path to output image.
  """
  # Copy the single channel if we are provided a grayscale image.
  if image.shape[2] == 1:
    image = np.repeat(image, 3, axis=2)
  image = np.ascontiguousarray(image)
  image *= 255.
  image = image.astype("uint8")
  with tf.gfile.Open(image_path, "wb") as path:
    img = Image.fromarray(image, mode="RGB")
    img.save(path)


def grid_save_images(images, image_path):
  """Saves images in list of [0,1]-valued np.arrays on a grid.

  Args:
    images: List of Numpy arrays of shape (height, width, {1,3}) with values in
      [0, 1].
    image_path: String with path to output image.
  """
  side_length = int(math.floor(math.sqrt(len(images))))
  image_rows = [
      np.concatenate(
          images[side_length * i:side_length * i + side_length], axis=0)
      for i in range(side_length)
  ]
  tiled_image = np.concatenate(image_rows, axis=1)
  save_image(tiled_image, image_path)


def padded_grid(images, num_rows=None, padding_px=10, value=None):
  """Creates a grid with padding in between images."""
  num_images = len(images)
  if num_rows is None:
    num_rows = best_num_rows(num_images)

  # Computes how many empty images we need to add.
  num_cols = int(np.ceil(float(num_images) / num_rows))
  num_missing = num_rows * num_cols - num_images

  # Add the empty images at the end.
  all_images = images + [np.ones_like(images[0])] * num_missing

  # Create the final grid.
  rows = [padded_stack(all_images[i * num_cols:(i + 1) * num_cols], padding_px,
                       1, value=value) for i in range(num_rows)]
  return padded_stack(rows, padding_px, axis=0, value=value)


def padded_stack(images, padding_px=10, axis=0, value=None):
  """Stacks images along axis with padding in between images."""
  padding_arr = padding_array(images[0], padding_px, axis, value=value)
  new_images = [images[0]]
  for image in images[1:]:
    new_images.append(padding_arr)
    new_images.append(image)
  return np.concatenate(new_images, axis=axis)


def padding_array(image, padding_px, axis, value=None):
  """Creates padding image of proper shape to pad image along the axis."""
  shape = list(image.shape)
  shape[axis] = padding_px
  if value is None:
    return np.ones(shape, dtype=image.dtype)
  else:
    assert len(value) == shape[-1]
    shape[-1] = 1
    return np.tile(value, shape)


def best_num_rows(num_elements, max_ratio=4):
  """Automatically selects a smart number of rows."""
  best_remainder = num_elements
  best_i = None
  i = int(np.sqrt(num_elements))
  while True:
    if num_elements > max_ratio * i * i:
      return best_i
    remainder = (i - num_elements % i) % i
    if remainder == 0:
      return i
    if remainder < best_remainder:
      best_remainder = remainder
      best_i = i
    i -= 1


def pad_around(image, padding_px=10, axis=None, value=None):
  """Adds a padding around each image."""
  # If axis is None, pad both the first and the second axis.
  if axis is None:
    image = pad_around(image, padding_px, axis=0, value=value)
    axis = 1
  padding_arr = padding_array(image, padding_px, axis, value=value)
  return np.concatenate([padding_arr, image, padding_arr], axis=axis)


def add_below(image, padding_px=10, value=None):
  """Adds a footer below."""
  if len(image.shape) == 2:
    image = np.expand_dims(image, -1)
  if image.shape[2] == 1:
    image = np.repeat(image, 3, 2)
  if image.shape[2] != 3:
    raise ValueError("Could not convert image to have three channels.")
  with tf.gfile.Open(resources.get_file("disentanglement_lib.png"), "rb") as f:
    footer = np.array(Image.open(f).convert("RGB")) * 1.0 / 255.
  missing_px = image.shape[1] - footer.shape[1]
  if missing_px < 0:
    return image
  if missing_px > 0:
    padding_arr = padding_array(footer, missing_px, axis=1, value=value)
    footer = np.concatenate([padding_arr, footer], axis=1)
  return padded_stack([image, footer], padding_px, axis=0, value=value)


def save_animation(list_of_animated_images, image_path, fps):
  full_size_images = []
  for single_images in zip(*list_of_animated_images):
    full_size_images.append(
        pad_around(add_below(padded_grid(list(single_images)))))
  imageio.mimwrite(image_path, full_size_images, fps=fps)


def cycle_factor(starting_index, num_indices, num_frames):
  """Cycles through the state space in a single cycle."""
  grid = np.linspace(starting_index, starting_index + 2*num_indices,
                     num=num_frames, endpoint=False)
  grid = np.array(np.ceil(grid), dtype=np.int64)
  grid -= np.maximum(0, 2*grid - 2*num_indices + 1)
  grid += np.maximum(0, -2*grid - 1)
  return grid


def cycle_gaussian(starting_value, num_frames, loc=0., scale=1.):
  """Cycles through the quantiles of a Gaussian in a single cycle."""
  starting_prob = scipy.stats.norm.cdf(starting_value, loc=loc, scale=scale)
  grid = np.linspace(starting_prob, starting_prob + 2.,
                     num=num_frames, endpoint=False)
  grid -= np.maximum(0, 2*grid - 2)
  grid += np.maximum(0, -2*grid)
  grid = np.minimum(grid, 0.999)
  grid = np.maximum(grid, 0.001)
  return np.array([scipy.stats.norm.ppf(i, loc=loc, scale=scale) for i in grid])


def cycle_interval(starting_value, num_frames, min_val, max_val):
  """Cycles through the state space in a single cycle."""
  starting_in_01 = (starting_value - min_val)/(max_val - min_val)
  grid = np.linspace(starting_in_01, starting_in_01 + 2.,
                     num=num_frames, endpoint=False)
  grid -= np.maximum(0, 2*grid - 2)
  grid += np.maximum(0, -2*grid)
  return grid * (max_val - min_val) + min_val



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility to access resources in package."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os


def get_file(path):
  """Returns path relative to file."""
  from pkg_resources import resource_filename  # pylint: disable=g-bad-import-order, g-import-not-at-top
  return resource_filename("disentanglement_lib", path)


def get_files_in_folder(path):
  import pkg_resources  # pylint: disable=g-bad-import-order, g-import-not-at-top
  for name in pkg_resources.resource_listdir("disentanglement_lib", path):
    new_path = path + "/" + name
    if not pkg_resources.resource_isdir("disentanglement_lib", new_path):
      yield pkg_resources.resource_filename("disentanglement_lib", new_path)





# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mig.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.evaluation.metrics import mig
import numpy as np
import gin.tf


def _identity_discretizer(target, num_bins):
  del num_bins
  return target


class MIGTest(absltest.TestCase):

  def test_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    representation_function = lambda x: x
    random_state = np.random.RandomState(0)
    scores = mig.compute_mig(
        ground_truth_data, representation_function, random_state, None, 3000)
    self.assertBetween(scores["discrete_mig"], 0.9, 1.0)

  def test_bad_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    representation_function = np.zeros_like
    random_state = np.random.RandomState(0)
    scores = mig.compute_mig(
        ground_truth_data, representation_function, random_state, None, 3000)
    self.assertBetween(scores["discrete_mig"], 0.0, 0.2)

  def test_duplicated_latent_space(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    def representation_function(x):
      x = np.array(x, dtype=np.float64)
      return np.hstack([x, x])
    random_state = np.random.RandomState(0)
    scores = mig.compute_mig(
        ground_truth_data, representation_function, random_state, None, 3000)
    self.assertBetween(scores["discrete_mig"], 0.0, 0.1)

if __name__ == "__main__":
  absltest.main()
